import existing_import
from existing_import import d
from m1 import A
from m2 import B
import m3
from m4 import D
from m5.sub import E
from m6 import F
from mStar import *
from m7 import a
from m8 import b
from ......m9 import c

def f(a1: A, a2: B, a3: m3.C, a4: D, a5: E, a6: F) -> G:
    pass

def g(a7: a, a8: b, a9: c, a10: d) -> existing_import:
    pass



from typing import Any

from braintree.search import Search as Search

class CustomerSearch:
    address_extended_address: Any
    address_first_name: Any
    address_last_name: Any
    address_locality: Any
    address_postal_code: Any
    address_region: Any
    address_street_address: Any
    address_country_name: Any
    cardholder_name: Any
    company: Any
    created_at: Any
    credit_card_expiration_date: Any
    credit_card_number: Any
    email: Any
    fax: Any
    first_name: Any
    id: Any
    ids: Any
    last_name: Any
    payment_method_token: Any
    payment_method_token_with_duplicates: Any
    phone: Any
    website: Any
    paypal_account_email: Any



from typing import Any

class Parser:
    BINDING_POWER: Any
    tokenizer: Any
    def __init__(self, lookahead: int = ...) -> None: ...
    def parse(self, expression): ...
    @classmethod
    def purge(cls) -> None: ...

class ParsedResult:
    expression: Any
    parsed: Any
    def __init__(self, expression, parsed) -> None: ...
    def search(self, value, options: Any | None = ...): ...



from datetime import date, datetime, time, timedelta

from MySQLdb._mysql import string_literal as string_literal

Date = date
Time = time
TimeDelta = timedelta
Timestamp = datetime
DateTimeDeltaType = timedelta
DateTimeType = datetime

def DateFromTicks(ticks): ...
def TimeFromTicks(ticks): ...
def TimestampFromTicks(ticks): ...

format_TIME = str
format_DATE = str

def format_TIMEDELTA(v): ...
def format_TIMESTAMP(d): ...
def DateTime_or_None(s): ...
def TimeDelta_or_None(s): ...
def Time_or_None(s): ...
def Date_or_None(s): ...
def DateTime2literal(d, c): ...
def DateTimeDelta2literal(d, c): ...



from typing import Any, Type

def assert_(condition: object) -> None: ...

ERR_FRAGMENT: str

class AuthorizationError(Exception):
    url: Any
    reason: str

class DAVError(Exception): ...
class PropsetError(DAVError): ...
class ProppatchError(DAVError): ...
class PropfindError(DAVError): ...
class ReportError(DAVError): ...
class MkcolError(DAVError): ...
class MkcalendarError(DAVError): ...
class PutError(DAVError): ...
class DeleteError(DAVError): ...
class NotFoundError(DAVError): ...
class ConsistencyError(DAVError): ...
class ReponseError(DAVError): ...

exception_by_method: dict[str, Type[DAVError]]



from typing import Any, Callable, Container, Iterable, List, Mapping, NamedTuple, Sequence, Union

LATEX_ESCAPE_RULES: dict[str, str]
MIN_PADDING: int
PRESERVE_WHITESPACE: bool
WIDE_CHARS_MODE: bool
multiline_formats: dict[str, str]
tabulate_formats: list[str]

class Line(NamedTuple):
    begin: str
    hline: str
    sep: str
    end: str

class DataRow(NamedTuple):
    begin: str
    sep: str
    end: str

_TableFormatLine = Union[None, Line, Callable[[List[int], List[str]], str]]
_TableFormatRow = Union[None, DataRow, Callable[[List[Any], List[int], List[str]], str]]

class TableFormat(NamedTuple):
    lineabove: _TableFormatLine
    linebelowheader: _TableFormatLine
    linebetweenrows: _TableFormatLine
    linebelow: _TableFormatLine
    headerrow: _TableFormatRow
    datarow: _TableFormatRow
    padding: int
    with_header_hide: Container[str] | None

def simple_separated_format(separator: str) -> TableFormat: ...
def tabulate(
    tabular_data: Mapping[str, Iterable[Any]] | Iterable[Iterable[Any]],
    headers: str | dict[str, str] | Sequence[str] = ...,
    tablefmt: str | TableFormat = ...,
    floatfmt: str | Iterable[str] = ...,
    numalign: str | None = ...,
    stralign: str | None = ...,
    missingval: str | Iterable[str] = ...,
    showindex: str | bool | Iterable[Any] = ...,
    disable_numparse: bool | Iterable[int] = ...,
    colalign: Iterable[str | None] | None = ...,
) -> str: ...



from _typeshed import WriteableBuffer
from io import BufferedIOBase, RawIOBase
from typing import Any, Callable, Protocol, Tuple, Type

BUFFER_SIZE: Any

class _Reader(Protocol):
    def read(self, __n: int) -> bytes: ...
    def seekable(self) -> bool: ...
    def seek(self, __n: int) -> Any: ...

class BaseStream(BufferedIOBase): ...

class DecompressReader(RawIOBase):
    def __init__(
        self,
        fp: _Reader,
        decomp_factory: Callable[..., object],
        trailing_error: Type[Exception] | Tuple[Type[Exception], ...] = ...,
        **decomp_args: Any,
    ) -> None: ...
    def readable(self) -> bool: ...
    def close(self) -> None: ...
    def seekable(self) -> bool: ...
    def readinto(self, b: WriteableBuffer) -> int: ...
    def read(self, size: int = ...) -> bytes: ...
    def seek(self, offset: int, whence: int = ...) -> int: ...
    def tell(self) -> int: ...



from typing import Any

from selenium.common.exceptions import (
    NoAlertPresentException as NoAlertPresentException,
    NoSuchElementException as NoSuchElementException,
    NoSuchFrameException as NoSuchFrameException,
    StaleElementReferenceException as StaleElementReferenceException,
    WebDriverException as WebDriverException,
)
from selenium.webdriver.remote.webdriver import WebElement as WebElement

class title_is:
    title: Any
    def __init__(self, title) -> None: ...
    def __call__(self, driver): ...

class title_contains:
    title: Any
    def __init__(self, title) -> None: ...
    def __call__(self, driver): ...

class presence_of_element_located:
    locator: Any
    def __init__(self, locator) -> None: ...
    def __call__(self, driver): ...

class url_contains:
    url: Any
    def __init__(self, url) -> None: ...
    def __call__(self, driver): ...

class url_matches:
    pattern: Any
    def __init__(self, pattern) -> None: ...
    def __call__(self, driver): ...

class url_to_be:
    url: Any
    def __init__(self, url) -> None: ...
    def __call__(self, driver): ...

class url_changes:
    url: Any
    def __init__(self, url) -> None: ...
    def __call__(self, driver): ...

class visibility_of_element_located:
    locator: Any
    def __init__(self, locator) -> None: ...
    def __call__(self, driver): ...

class visibility_of:
    element: Any
    def __init__(self, element) -> None: ...
    def __call__(self, ignored): ...

class presence_of_all_elements_located:
    locator: Any
    def __init__(self, locator) -> None: ...
    def __call__(self, driver): ...

class visibility_of_any_elements_located:
    locator: Any
    def __init__(self, locator) -> None: ...
    def __call__(self, driver): ...

class visibility_of_all_elements_located:
    locator: Any
    def __init__(self, locator) -> None: ...
    def __call__(self, driver): ...

class text_to_be_present_in_element:
    locator: Any
    text: Any
    def __init__(self, locator, text_) -> None: ...
    def __call__(self, driver): ...

class text_to_be_present_in_element_value:
    locator: Any
    text: Any
    def __init__(self, locator, text_) -> None: ...
    def __call__(self, driver): ...

class frame_to_be_available_and_switch_to_it:
    frame_locator: Any
    def __init__(self, locator) -> None: ...
    def __call__(self, driver): ...

class invisibility_of_element_located:
    target: Any
    def __init__(self, locator) -> None: ...
    def __call__(self, driver): ...

class invisibility_of_element(invisibility_of_element_located): ...

class element_to_be_clickable:
    locator: Any
    def __init__(self, locator) -> None: ...
    def __call__(self, driver): ...

class staleness_of:
    element: Any
    def __init__(self, element) -> None: ...
    def __call__(self, ignored): ...

class element_to_be_selected:
    element: Any
    def __init__(self, element) -> None: ...
    def __call__(self, ignored): ...

class element_located_to_be_selected:
    locator: Any
    def __init__(self, locator) -> None: ...
    def __call__(self, driver): ...

class element_selection_state_to_be:
    element: Any
    is_selected: Any
    def __init__(self, element, is_selected) -> None: ...
    def __call__(self, ignored): ...

class element_located_selection_state_to_be:
    locator: Any
    is_selected: Any
    def __init__(self, locator, is_selected) -> None: ...
    def __call__(self, driver): ...

class number_of_windows_to_be:
    num_windows: Any
    def __init__(self, num_windows) -> None: ...
    def __call__(self, driver): ...

class new_window_is_opened:
    current_handles: Any
    def __init__(self, current_handles) -> None: ...
    def __call__(self, driver): ...

class alert_is_present:
    def __init__(self) -> None: ...
    def __call__(self, driver): ...



import abc
from typing import Any

ConnectionError: Any

class GlobalCache(metaclass=abc.ABCMeta):
    __metaclass__: Any
    transient_errors: Any
    strict_read: bool
    strict_write: bool
    @abc.abstractmethod
    def get(self, keys): ...
    @abc.abstractmethod
    def set(self, items, expires: Any | None = ...): ...
    @abc.abstractmethod
    def delete(self, keys): ...
    @abc.abstractmethod
    def watch(self, keys): ...
    @abc.abstractmethod
    def unwatch(self, keys): ...
    @abc.abstractmethod
    def compare_and_swap(self, items, expires: Any | None = ...): ...
    @abc.abstractmethod
    def clear(self): ...

class _InProcessGlobalCache(GlobalCache):
    cache: Any
    def __init__(self) -> None: ...
    def get(self, keys): ...
    def set(self, items, expires: Any | None = ...) -> None: ...
    def delete(self, keys) -> None: ...
    def watch(self, keys) -> None: ...
    def unwatch(self, keys) -> None: ...
    def compare_and_swap(self, items, expires: Any | None = ...): ...
    def clear(self) -> None: ...

class RedisCache(GlobalCache):
    transient_errors: Any
    @classmethod
    def from_environment(cls, strict_read: bool = ..., strict_write: bool = ...): ...
    redis: Any
    strict_read: Any
    strict_write: Any
    def __init__(self, redis, strict_read: bool = ..., strict_write: bool = ...) -> None: ...
    @property
    def pipes(self): ...
    def get(self, keys): ...
    def set(self, items, expires: Any | None = ...) -> None: ...
    def delete(self, keys) -> None: ...
    def watch(self, keys) -> None: ...
    def unwatch(self, keys) -> None: ...
    def compare_and_swap(self, items, expires: Any | None = ...): ...
    def clear(self) -> None: ...

class MemcacheCache(GlobalCache):
    class KeyNotSet(Exception):
        key: Any
        def __init__(self, key) -> None: ...
        def __eq__(self, other): ...
    transient_errors: Any
    @classmethod
    def from_environment(cls, max_pool_size: int = ..., strict_read: bool = ..., strict_write: bool = ...): ...
    client: Any
    strict_read: Any
    strict_write: Any
    def __init__(self, client, strict_read: bool = ..., strict_write: bool = ...) -> None: ...
    @property
    def caskeys(self): ...
    def get(self, keys): ...
    def set(self, items, expires: Any | None = ...): ...
    def delete(self, keys) -> None: ...
    def watch(self, keys) -> None: ...
    def unwatch(self, keys) -> None: ...
    def compare_and_swap(self, items, expires: Any | None = ...): ...
    def clear(self) -> None: ...



from typing import Any

class Plugin:
    capability: Any
    @classmethod
    def is_capable(cls, requested_capability): ...

def get_plugin(cls, requested_capability: Any | None = ...): ...
def load_plugins(config): ...



import sys
import types
from opcode import (
    EXTENDED_ARG as EXTENDED_ARG,
    HAVE_ARGUMENT as HAVE_ARGUMENT,
    cmp_op as cmp_op,
    hascompare as hascompare,
    hasconst as hasconst,
    hasfree as hasfree,
    hasjabs as hasjabs,
    hasjrel as hasjrel,
    haslocal as haslocal,
    hasname as hasname,
    hasnargs as hasnargs,
    opmap as opmap,
    opname as opname,
    stack_effect as stack_effect,
)
from typing import IO, Any, Callable, Iterator, NamedTuple, Union

# Strictly this should not have to include Callable, but mypy doesn't use FunctionType
# for functions (python/mypy#3171)
_have_code = Union[types.MethodType, types.FunctionType, types.CodeType, type, Callable[..., Any]]
_have_code_or_string = Union[_have_code, str, bytes]

class Instruction(NamedTuple):
    opname: str
    opcode: int
    arg: int | None
    argval: Any
    argrepr: str
    offset: int
    starts_line: int | None
    is_jump_target: bool

class Bytecode:
    codeobj: types.CodeType
    first_line: int
    def __init__(self, x: _have_code_or_string, *, first_line: int | None = ..., current_offset: int | None = ...) -> None: ...
    def __iter__(self) -> Iterator[Instruction]: ...
    def __repr__(self) -> str: ...
    def info(self) -> str: ...
    def dis(self) -> str: ...
    @classmethod
    def from_traceback(cls, tb: types.TracebackType) -> Bytecode: ...

COMPILER_FLAG_NAMES: dict[int, str]

def findlabels(code: _have_code) -> list[int]: ...
def findlinestarts(code: _have_code) -> Iterator[tuple[int, int]]: ...
def pretty_flags(flags: int) -> str: ...
def code_info(x: _have_code_or_string) -> str: ...

if sys.version_info >= (3, 7):
    def dis(x: _have_code_or_string | None = ..., *, file: IO[str] | None = ..., depth: int | None = ...) -> None: ...

else:
    def dis(x: _have_code_or_string | None = ..., *, file: IO[str] | None = ...) -> None: ...

def distb(tb: types.TracebackType | None = ..., *, file: IO[str] | None = ...) -> None: ...
def disassemble(co: _have_code, lasti: int = ..., *, file: IO[str] | None = ...) -> None: ...
def disco(co: _have_code, lasti: int = ..., *, file: IO[str] | None = ...) -> None: ...
def show_code(co: _have_code, *, file: IO[str] | None = ...) -> None: ...
def get_instructions(x: _have_code, *, first_line: int | None = ...) -> Iterator[Instruction]: ...



from typing import Text

# Python 2 accepts unicode ascii pretty much everywhere.
_Bytes = Text
_Ascii = Text

def a2b_uu(__data: _Ascii) -> bytes: ...
def b2a_uu(__data: _Bytes) -> bytes: ...
def a2b_base64(__data: _Ascii) -> bytes: ...
def b2a_base64(__data: _Bytes) -> bytes: ...
def a2b_qp(data: _Ascii, header: bool = ...) -> bytes: ...
def b2a_qp(data: _Bytes, quotetabs: bool = ..., istext: bool = ..., header: bool = ...) -> bytes: ...
def a2b_hqx(__data: _Ascii) -> bytes: ...
def rledecode_hqx(__data: _Bytes) -> bytes: ...
def rlecode_hqx(__data: _Bytes) -> bytes: ...
def b2a_hqx(__data: _Bytes) -> bytes: ...
def crc_hqx(__data: _Bytes, __crc: int) -> int: ...
def crc32(__data: _Bytes, __crc: int = ...) -> int: ...
def b2a_hex(__data: _Bytes) -> bytes: ...
def hexlify(__data: _Bytes) -> bytes: ...
def a2b_hex(__hexstr: _Ascii) -> bytes: ...
def unhexlify(__hexstr: _Ascii) -> bytes: ...

class Error(ValueError): ...
class Incomplete(Exception): ...



from typing import Any, Callable, NoReturn, Tuple

TIMEOUT_MAX: int
error = RuntimeError

def start_new_thread(function: Callable[..., Any], args: Tuple[Any, ...], kwargs: dict[str, Any] = ...) -> None: ...
def exit() -> NoReturn: ...
def get_ident() -> int: ...
def allocate_lock() -> LockType: ...
def stack_size(size: int | None = ...) -> int: ...

class LockType(object):
    locked_status: bool
    def __init__(self) -> None: ...
    def acquire(self, waitflag: bool | None = ..., timeout: int = ...) -> bool: ...
    def __enter__(self, waitflag: bool | None = ..., timeout: int = ...) -> bool: ...
    def __exit__(self, typ: Any, val: Any, tb: Any) -> None: ...
    def release(self) -> bool: ...
    def locked(self) -> bool: ...

def interrupt_main() -> None: ...



from _typeshed import SupportsWrite
from collections.abc import Iterable
from typing import Any, Iterator, TypeVar, overload
from typing_extensions import Literal

DEBUG: bool
CR: str
LF: str
CRLF: str
SPACE: str
TAB: str
SPACEORTAB: str

_V = TypeVar("_V", bound=VBase)
_W = TypeVar("_W", bound=SupportsWrite[bytes])

class VBase:
    group: Any | None
    behavior: Any | None
    parentBehavior: Any | None
    isNative: bool
    def __init__(self, group: Any | None = ...) -> None: ...
    def copy(self, copyit: VBase) -> None: ...
    def validate(self, *args, **kwds) -> bool: ...
    def getChildren(self) -> list[Any]: ...
    def clearBehavior(self, cascade: bool = ...) -> None: ...
    def autoBehavior(self, cascade: bool = ...) -> None: ...
    def setBehavior(self, behavior, cascade: bool = ...) -> None: ...
    def transformToNative(self): ...
    def transformFromNative(self): ...
    def transformChildrenToNative(self) -> None: ...
    def transformChildrenFromNative(self, clearBehavior: bool = ...) -> None: ...
    @overload
    def serialize(self, buf: None = ..., lineLength: int = ..., validate: bool = ..., behavior: Any | None = ...) -> str: ...
    @overload
    def serialize(self, buf: _W, lineLength: int = ..., validate: bool = ..., behavior: Any | None = ...) -> _W: ...

def toVName(name, stripNum: int = ..., upper: bool = ...): ...

class ContentLine(VBase):
    name: Any
    encoded: Any
    params: Any
    singletonparams: Any
    isNative: Any
    lineNumber: Any
    value: Any
    def __init__(
        self,
        name,
        params,
        value,
        group: Any | None = ...,
        encoded: bool = ...,
        isNative: bool = ...,
        lineNumber: Any | None = ...,
        *args,
        **kwds,
    ) -> None: ...
    @classmethod
    def duplicate(cls, copyit): ...
    def copy(self, copyit) -> None: ...
    def __eq__(self, other): ...
    def __getattr__(self, name): ...
    def __setattr__(self, name, value) -> None: ...
    def __delattr__(self, name) -> None: ...
    def valueRepr(self): ...
    def __unicode__(self): ...
    def prettyPrint(self, level: int = ..., tabwidth: int = ...) -> None: ...

class Component(VBase):
    contents: dict[str, list[VBase]]
    name: Any
    useBegin: bool
    def __init__(self, name: Any | None = ..., *args, **kwds) -> None: ...
    @classmethod
    def duplicate(cls, copyit): ...
    def copy(self, copyit) -> None: ...
    def setProfile(self, name) -> None: ...
    def __getattr__(self, name): ...
    normal_attributes: Any
    def __setattr__(self, name, value) -> None: ...
    def __delattr__(self, name) -> None: ...
    def getChildValue(self, childName, default: Any | None = ..., childNumber: int = ...): ...
    @overload
    def add(self, objOrName: _V, group: str | None = ...) -> _V: ...
    @overload
    def add(self, objOrName: Literal["vevent"], group: str | None = ...) -> Component: ...
    @overload
    def add(
        self, objOrName: Literal["uid", "summary", "description", "dtstart", "dtend"], group: str | None = ...
    ) -> ContentLine: ...
    @overload
    def add(self, objOrName: str, group: str | None = ...) -> Any: ...  # returns VBase sub-class
    def remove(self, obj) -> None: ...
    def getChildren(self) -> list[Any]: ...
    def components(self) -> Iterable[Component]: ...
    def lines(self): ...
    def sortChildKeys(self): ...
    def getSortedChildren(self): ...
    def setBehaviorFromVersionLine(self, versionLine) -> None: ...
    def transformChildrenToNative(self) -> None: ...
    def transformChildrenFromNative(self, clearBehavior: bool = ...) -> None: ...
    def prettyPrint(self, level: int = ..., tabwidth: int = ...) -> None: ...

class VObjectError(Exception):
    msg: Any
    lineNumber: Any
    def __init__(self, msg, lineNumber: Any | None = ...) -> None: ...

class ParseError(VObjectError): ...
class ValidateError(VObjectError): ...
class NativeError(VObjectError): ...

patterns: Any
param_values_re: Any
params_re: Any
line_re: Any
begin_re: Any

def parseParams(string): ...
def parseLine(line, lineNumber: Any | None = ...): ...

wrap_re: Any
logical_lines_re: Any
testLines: str

def getLogicalLines(fp, allowQP: bool = ...) -> None: ...
def textLineToContentLine(text, n: Any | None = ...): ...
def dquoteEscape(param): ...
def foldOneLine(outbuf, input, lineLength: int = ...) -> None: ...
def defaultSerialize(obj, buf, lineLength): ...

class Stack:
    stack: Any
    def __init__(self) -> None: ...
    def __len__(self): ...
    def top(self): ...
    def topName(self): ...
    def modifyTop(self, item) -> None: ...
    def push(self, obj) -> None: ...
    def pop(self): ...

def readComponents(
    streamOrString, validate: bool = ..., transform: bool = ..., ignoreUnreadable: bool = ..., allowQP: bool = ...
) -> Iterator[Component]: ...
def readOne(stream, validate: bool = ..., transform: bool = ..., ignoreUnreadable: bool = ..., allowQP: bool = ...): ...
def registerBehavior(behavior, name: Any | None = ..., default: bool = ..., id: Any | None = ...) -> None: ...
def getBehavior(name, id: Any | None = ...): ...
def newFromBehavior(name, id: Any | None = ...): ...
def backslashEscape(s): ...



"""Tests for booleq.py."""

from pytype.pytd import booleq
import unittest


# pylint: disable=invalid-name
And = booleq.And
Or = booleq.Or
Eq = booleq.Eq
TRUE = booleq.TRUE
FALSE = booleq.FALSE


class TestBoolEq(unittest.TestCase):
  """Test algorithms and datastructures of booleq.py."""

  def test_true_and_false(self):
    self.assertNotEqual(TRUE, FALSE)
    self.assertNotEqual(FALSE, TRUE)
    self.assertEqual(TRUE, TRUE)
    self.assertEqual(FALSE, FALSE)

  def test_equality(self):
    self.assertEqual(Eq("a", "b"), Eq("b", "a"))
    self.assertEqual(Eq("a", "b"), Eq("a", "b"))
    self.assertNotEqual(Eq("a", "a"), Eq("a", "b"))
    self.assertNotEqual(Eq("b", "a"), Eq("b", "b"))

  def test_and(self):
    self.assertEqual(TRUE, And([]))
    self.assertEqual(TRUE, And([TRUE]))
    self.assertEqual(TRUE, And([TRUE, TRUE]))
    self.assertEqual(FALSE, And([TRUE, FALSE]))
    self.assertEqual(Eq("a", "b"), And([Eq("a", "b"), TRUE]))
    self.assertEqual(FALSE, And([Eq("a", "b"), FALSE]))

  def test_or(self):
    self.assertEqual(FALSE, Or([]))
    self.assertEqual(TRUE, Or([TRUE]))
    self.assertEqual(TRUE, Or([TRUE, TRUE]))
    self.assertEqual(TRUE, Or([TRUE, FALSE]))
    self.assertEqual(Eq("a", "b"), Or([Eq("a", "b"), FALSE]))
    self.assertEqual(TRUE, Or([Eq("a", "b"), TRUE]))

  def test_nested_equals(self):
    eq1 = Eq("a", "u")
    eq2 = Eq("b", "v")
    eq3 = Eq("c", "w")
    eq4 = Eq("d", "x")
    nested = Or([And([eq1, eq2]), And([eq3, eq4])])
    self.assertEqual(nested, nested)

  def test_order(self):
    eq1 = Eq("a", "b")
    eq2 = Eq("b", "c")
    self.assertEqual(Or([eq1, eq2]), Or([eq2, eq1]))
    self.assertEqual(And([eq1, eq2]), And([eq2, eq1]))

  def test_hash(self):
    eq1 = Eq("a", "b")
    eq2 = Eq("b", "c")
    eq3 = Eq("c", "d")
    self.assertEqual(hash(Eq("x", "y")), hash(Eq("y", "x")))
    self.assertEqual(hash(Or([eq1, eq2, eq3])), hash(Or([eq2, eq3, eq1])))
    self.assertEqual(hash(And([eq1, eq2, eq3])), hash(And([eq2, eq3, eq1])))

  def test_pivots(self):
    values = {"x": {"0", "1"},
              "y": {"0", "1"}}
    # x == 0 || x == 1
    equation = Or([Eq("x", "0"), Eq("x", "1")])
    self.assertCountEqual(["0", "1"], equation.extract_pivots(values)["x"])

    # x == 0 && x == 0
    equation = And([Eq("x", "0"), Eq("x", "0")])
    self.assertCountEqual(["0"], equation.extract_pivots(values)["x"])

    # x == 0 && (x == 0 || x == 1)
    equation = And([Eq("x", "0"), Or([Eq("x", "0"), Eq("x", "1")])])
    self.assertCountEqual(["0"], equation.extract_pivots(values)["x"])

    # x == 0 || x == 0
    equation = And([Eq("x", "0"), Eq("x", "0")])
    self.assertCountEqual(["0"], equation.extract_pivots(values)["x"])

    # x == 0 || y == 0
    equation = Or([Eq("x", "0"), Eq("y", "0")])
    pivots = equation.extract_pivots(values)
    self.assertCountEqual(["0"], pivots["x"])
    self.assertCountEqual(["0"], pivots["y"])

  def test_simplify(self):
    # x == 0 || x == 1  with x in {0}
    equation = Or([Eq("x", "0"), Eq("x", "1")])
    values = {"x": {"0"}}
    self.assertEqual(Eq("x", "0"), equation.simplify(values))

    # x == 0 || x == 1  with x in {0}
    equation = Or([Eq("x", "0"), Eq("x", "1")])
    values = {"x": {"0", "1"}}
    self.assertEqual(equation, equation.simplify(values))

    # x == 0 with x in {1}
    equation = Eq("x", "0")
    values = {"x": {"1"}}
    self.assertEqual(FALSE, equation.simplify(values))

    # x == 0 with x in {0}
    equation = Eq("x", "0")
    values = {"x": {"0"}}
    self.assertEqual(equation, equation.simplify(values))

    # x == 0 && y == 0 with x in {1}, y in {1}
    equation = Or([Eq("x", "0"), Eq("y", "1")])
    values = {"x": {"1"}, "y": {"1"}}
    self.assertEqual(Eq("y", "1"), equation.simplify(values))

    # x == 0 && y == 0 with x in {0}, y in {1}
    equation = Or([Eq("x", "0"), Eq("y", "1")])
    values = {"x": {"0"}, "y": {"1"}}
    self.assertEqual(equation, equation.simplify(values))

    # x == 0 && x == 0 with x in {0}
    equation = And([Eq("x", "0"), Eq("x", "0")])
    values = {"x": {"0"}}
    self.assertEqual(Eq("x", "0"), equation.simplify(values))

    # x == y with x in {0, 1} and y in {1, 2}
    equation = Eq("x", "y")
    values = {"x": {"0", "1"}, "y": {"1", "2"}}
    self.assertEqual(equation, equation.simplify(values))

  def _MakeSolver(self, variables=("x", "y")):
    solver = booleq.Solver()
    for variable in variables:
      solver.register_variable(variable)
    return solver

  def test_get_false_first_approximation(self):
    solver = self._MakeSolver(["x"])
    solver.implies(Eq("x", "1"), FALSE)
    self.assertDictEqual(solver._get_first_approximation(), {"x": set()})

  def test_get_unrelated_first_approximation(self):
    solver = self._MakeSolver()
    solver.implies(Eq("x", "1"), TRUE)
    solver.implies(Eq("y", "2"), TRUE)
    self.assertDictEqual(solver._get_first_approximation(),
                         {"x": {"1"}, "y": {"2"}})

  def test_get_equal_first_approximation(self):
    solver = self._MakeSolver()
    solver.implies(Eq("x", "1"), Eq("x", "y"))
    assignments = solver._get_first_approximation()
    self.assertDictEqual(assignments,
                         {"x": {"1"}, "y": {"1"}})
    self.assertIs(assignments["x"], assignments["y"])

  def test_get_multiple_equal_first_approximation(self):
    solver = self._MakeSolver(["x", "y", "z"])
    solver.implies(Eq("y", "1"), Eq("x", "y"))
    solver.implies(Eq("z", "2"), Eq("y", "z"))
    assignments = solver._get_first_approximation()
    self.assertDictEqual(assignments,
                         {"x": {"1", "2"},
                          "y": {"1", "2"},
                          "z": {"1", "2"}})
    self.assertIs(assignments["x"], assignments["y"])
    self.assertIs(assignments["y"], assignments["z"])

  def test_implication(self):
    solver = self._MakeSolver()
    solver.implies(Eq("x", "1"), Eq("y", "1"))
    solver.implies(Eq("x", "2"), FALSE)  # not Eq("x", "2")
    self.assertDictEqual(solver.solve(),
                         {"x": {"1"},
                          "y": {"1"}})

  def test_ground_truth(self):
    solver = self._MakeSolver()
    solver.implies(Eq("x", "1"), Eq("y", "1"))
    solver.always_true(Eq("x", "1"))
    self.assertDictEqual(solver.solve(),
                         {"x": {"1"},
                          "y": {"1"}})

  def test_filter(self):
    solver = self._MakeSolver(["x", "y"])
    solver.implies(Eq("x", "1"), TRUE)
    solver.implies(Eq("x", "2"), FALSE)
    solver.implies(Eq("x", "3"), FALSE)
    solver.implies(Eq("y", "1"), Or([Eq("x", "1"), Eq("x", "2"), Eq("x", "3")]))
    solver.implies(Eq("y", "2"), Or([Eq("x", "2"), Eq("x", "3")]))
    solver.implies(Eq("y", "3"), Or([Eq("x", "2")]))
    self.assertDictEqual(solver.solve(),
                         {"x": {"1"},
                          "y": {"1"}})

  def test_solve_and(self):
    solver = self._MakeSolver(["x", "y", "z"])
    solver.always_true(Eq("x", "1"))
    solver.implies(Eq("y", "1"), And([Eq("x", "1"), Eq("z", "1")]))
    solver.implies(Eq("x", "1"), And([Eq("y", "1"), Eq("z", "1")]))
    solver.implies(Eq("z", "1"), And([Eq("x", "1"), Eq("y", "1")]))
    self.assertDictEqual(solver.solve(),
                         {"x": {"1"},
                          "y": {"1"},
                          "z": {"1"}})

  def test_solve_twice(self):
    solver = self._MakeSolver()
    solver.implies(Eq("x", "1"), Or([Eq("y", "1"), Eq("y", "2")]))
    solver.implies(Eq("y", "1"), FALSE)
    self.assertDictEqual(solver.solve(), solver.solve())

  def test_change_after_solve(self):
    solver = self._MakeSolver()
    solver.solve()
    self.assertRaises(AssertionError, solver.register_variable, "z")
    self.assertRaises(AssertionError, solver.implies, Eq("x", "1"), TRUE)

  def test_nested(self):
    solver = booleq.Solver()
    solver.register_variable("x")
    solver.register_variable("y")
    solver.register_variable("z")
    solver.implies(Eq("x", "b"), Eq("y", "b"))
    solver.implies(Eq("x", "d"), Eq("y", "z"))
    solver.implies(Eq("x", "e"), Eq("y", "e"))
    solver.implies(Eq("y", "a"), TRUE)
    solver.implies(Eq("y", "b"), TRUE)
    solver.implies(Eq("y", "d"), FALSE)
    solver.implies(Eq("y", "e"), FALSE)
    m = solver.solve()
    self.assertCountEqual(m["z"], {"a", "b"})

  def test_conjunction(self):
    solver = booleq.Solver()
    solver.register_variable("x")
    solver.register_variable("y")
    solver.register_variable("y.T")
    solver.register_variable("z")
    solver.register_variable("z.T")
    solver.register_variable("w")
    solver.implies(Eq("x", "1"), And([Eq("y", "2"), Eq("y.T", "1")]))
    solver.implies(Eq("y", "2"), And([Eq("z", "3"), Eq("z.T", "y.T")]))
    solver.implies(Eq("z", "3"),
                   Eq("w", "z.T"))
    solver.implies(Eq("w", "1"), TRUE)
    solver.implies(Eq("w", "4"), TRUE)
    m = solver.solve()
    self.assertCountEqual(m["x"], {"1"})
    self.assertCountEqual(m["y"], {"2"})
    self.assertCountEqual(m["z"], {"3"})
    self.assertCountEqual(m["z.T"], {"1"})
    self.assertIn("1", m["y.T"])
    self.assertNotIn("4", m["y.T"])

if __name__ == "__main__":
  unittest.main()



"""Tests for abstract.py."""

from pytype import config
from pytype import context
from pytype import errors
from pytype import load_pytd
from pytype import special_builtins
from pytype import state as frame_state
from pytype.abstract import abstract
from pytype.abstract import abstract_utils
from pytype.abstract import function
from pytype.pytd import pytd
from pytype.pytd import pytd_utils
from pytype.tests import test_base
from pytype.typegraph import cfg

import unittest


class AbstractTestBase(test_base.UnitTest):

  def setUp(self):
    super().setUp()
    options = config.Options.create(python_version=self.python_version)
    self._ctx = context.Context(errors.ErrorLog(), options,
                                load_pytd.Loader(options))
    self._program = self._ctx.program
    self._node = self._ctx.root_node.ConnectNew("test_node")

  def new_var(self, *values):
    """Create a Variable bound to the given values."""
    var = self._program.NewVariable()
    for value in values:
      var.AddBinding(value, source_set=(), where=self._node)
    return var

  def new_dict(self, **kwargs):
    """Create a Dict from keywords mapping names to Variable objects."""
    d = abstract.Dict(self._ctx)
    for name, var in kwargs.items():
      d.set_str_item(self._node, name, var)
    return d


class IsInstanceTest(AbstractTestBase):

  def setUp(self):
    super().setUp()
    self._is_instance = special_builtins.IsInstance.make(self._ctx)
    # Easier access to some primitive instances.
    self._bool = self._ctx.convert.primitive_class_instances[bool]
    self._int = self._ctx.convert.primitive_class_instances[int]
    self._str = self._ctx.convert.primitive_class_instances[str]
    # Values that represent primitive classes.
    self._obj_class = self._ctx.convert.primitive_classes[object]
    self._int_class = self._ctx.convert.primitive_classes[int]
    self._str_class = self._ctx.convert.primitive_classes[str]

  def assert_call(self, expected, left, right):
    """Check that call() returned the desired results.

    Args:
      expected: A dict from values to source sets, where a source set is
          represented by the sorted binding names separated by spaces, for
          example "left:0 right:1" would indicate binding #0 of variable
          "left" and binding #1 of variable "right".
      left: A Variable to use as the first arg to call().
      right: A Variable to use as the second arg to call().
    """
    name_map = {left: "left", right: "right"}
    node, result = self._is_instance.call(
        self._node, None, function.Args(
            (left, right), self.new_dict(), None, None))
    self.assertIn(node, self._node.outgoing)
    result_map = {}
    # Turning source sets into canonical string representations of the binding
    # names makes it much easier to debug failures.
    for b in result.bindings:
      terms = set()
      for o in b.origins:
        self.assertEqual(node, o.where)
        for sources in o.source_sets:
          terms.add(" ".join(sorted(
              "%s:%d" % (name_map[b.variable], b.variable.bindings.index(b))
              for b in sources)))
      result_map[b.data] = terms
    self.assertEqual(expected, result_map)

  def test_call_single_bindings(self):
    right = self.new_var(self._str_class)
    left = self.new_var(self._str)
    self.assert_call({self._ctx.convert.true: {"left:0 right:0"}}, left, right)
    left = self.new_var(self._int)
    self.assert_call({self._ctx.convert.false: {"left:0 right:0"}}, left, right)
    left = self.new_var(abstract.Unknown(self._ctx))
    self.assert_call(
        {self._bool: {"left:0 right:0"}},
        left, right)

  def test_call_multiple_bindings(self):
    left = self.new_var(self._int, self._str)
    right = self.new_var(self._int_class, self._str_class)
    self.assert_call(
        {
            self._ctx.convert.true: {"left:0 right:0", "left:1 right:1"},
            self._ctx.convert.false: {"left:0 right:1", "left:1 right:0"},
        }, left, right)

  def test_call_wrong_argcount(self):
    self._ctx.vm.push_frame(frame_state.SimpleFrame())
    node, result = self._is_instance.call(
        self._node, None, function.Args((), self.new_dict(), None, None))
    self.assertEqual(self._node, node)
    self.assertIsInstance(abstract_utils.get_atomic_value(result),
                          abstract.Unsolvable)
    self.assertRegex(str(self._ctx.errorlog), "missing-parameter")

  def test_call_wrong_keywords(self):
    self._ctx.vm.push_frame(frame_state.SimpleFrame())
    x = self.new_var(abstract.Unknown(self._ctx))
    node, result = self._is_instance.call(
        self._node, None, function.Args(
            (x, x), self.new_dict(foo=x), None, None))
    self.assertEqual(self._node, node)
    self.assertIsInstance(abstract_utils.get_atomic_value(result),
                          abstract.Unsolvable)
    self.assertRegex(
        str(self._ctx.errorlog), r"foo.*isinstance.*\[wrong-keyword-args\]")

  def test_is_instance(self):
    def check(expected, left, right):
      self.assertEqual(expected, self._is_instance._is_instance(left, right))

    # Unknown and Unsolvable are ambiguous.
    check(None, abstract.Unknown(self._ctx), self._obj_class)
    check(None, abstract.Unsolvable(self._ctx), self._obj_class)

    # If the object's class has multiple bindings, result is ambiguous.
    obj = abstract.SimpleValue("foo", self._ctx)
    check(None, obj, self._obj_class)
    obj.set_class(self._node, self.new_var(
        self._str_class, self._int_class))
    check(None, obj, self._str_class)

    # If the class_spec is not a class, result is ambiguous.
    check(None, self._str, self._str)

    # Result is True/False depending on if the class is in the object's mro.
    check(True, self._str, self._obj_class)
    check(True, self._str, self._str_class)
    check(False, self._str, self._int_class)

  def test_flatten(self):
    def maybe_var(v):
      return v if isinstance(v, cfg.Variable) else self.new_var(v)

    def new_tuple(*args):
      pyval = tuple(maybe_var(a) for a in args)
      return self._ctx.convert.tuple_to_value(pyval)

    def check(expected_ambiguous, expected_classes, value):
      classes = []
      ambiguous = abstract_utils.flatten(value, classes)
      self.assertEqual(expected_ambiguous, ambiguous)
      self.assertEqual(expected_classes, classes)

    unknown = abstract.Unknown(self._ctx)

    # Simple values.
    check(False, [self._str_class], self._str_class)
    check(True, [], self._str)
    check(True, [], unknown)

    # (str, int)
    check(False, [self._str_class, self._int_class],
          new_tuple(self._str_class, self._int_class))
    # (str, ?, int)
    check(True, [self._str_class, self._int_class],
          new_tuple(self._str_class, unknown, self._int_class))
    # (str, (int, object))
    check(False, [self._str_class, self._int_class, self._obj_class],
          new_tuple(
              self._str_class,
              new_tuple(self._int_class, self._obj_class)))
    # (str, (?, object))
    check(True, [self._str_class, self._obj_class],
          new_tuple(
              self._str_class,
              new_tuple(unknown, self._obj_class)))
    # A variable with multiple bindings is ambiguous.
    # (str, int | object)
    check(True, [self._str_class],
          new_tuple(self._str_class,
                    self.new_var(self._int_class, self._obj_class)))


class PyTDTest(AbstractTestBase):
  """Tests for abstract -> pytd type conversions."""

  def test_metaclass(self):
    cls = abstract.InterpreterClass("X", [], {}, None, self._ctx)
    meta = abstract.InterpreterClass("M", [], {}, None, self._ctx)
    meta.official_name = "M"
    cls.cls = meta
    pytd_cls = cls.to_pytd_def(self._ctx.root_node, "X")
    self.assertEqual(pytd_cls.metaclass, pytd.NamedType("M"))

  def test_inherited_metaclass(self):
    base = abstract.InterpreterClass("X", [], {}, None, self._ctx)
    base.official_name = "X"
    meta = abstract.InterpreterClass("M", [], {}, None, self._ctx)
    meta.official_name = "M"
    base.cls = meta
    child = abstract.InterpreterClass("Y",
                                      [base.to_variable(self._ctx.root_node)],
                                      {}, None, self._ctx)
    self.assertIs(child.cls, base.cls)
    pytd_cls = child.to_pytd_def(self._ctx.root_node, "Y")
    self.assertIs(pytd_cls.metaclass, None)

  def test_metaclass_union(self):
    cls = abstract.InterpreterClass("X", [], {}, None, self._ctx)
    meta1 = abstract.InterpreterClass("M1", [], {}, None, self._ctx)
    meta2 = abstract.InterpreterClass("M2", [], {}, None, self._ctx)
    meta1.official_name = "M1"
    meta2.official_name = "M2"
    cls.cls = abstract.Union([meta1, meta2], self._ctx)
    pytd_cls = cls.to_pytd_def(self._ctx.root_node, "X")
    self.assertEqual(pytd_cls.metaclass, pytd.UnionType(
        (pytd.NamedType("M1"), pytd.NamedType("M2"))))

  def test_to_type_with_view1(self):
    # to_type(<instance of List[int or unsolvable]>, view={T: int})
    instance = abstract.List([], self._ctx)
    instance.merge_instance_type_parameter(
        self._ctx.root_node, abstract_utils.T,
        self._ctx.program.NewVariable([self._ctx.convert.unsolvable], [],
                                      self._ctx.root_node))
    param_binding = instance.get_instance_type_parameter(
        abstract_utils.T).AddBinding(
            self._ctx.convert.primitive_class_instances[int], [],
            self._ctx.root_node)
    view = {
        instance.get_instance_type_parameter(abstract_utils.T): param_binding}
    pytd_type = instance.to_type(self._ctx.root_node, seen=None, view=view)
    self.assertEqual("builtins.list", pytd_type.name)
    self.assertSetEqual({"builtins.int"},
                        {t.name for t in pytd_type.parameters})

  def test_to_type_with_view2(self):
    # to_type(<tuple (int or str,)>, view={0: str})
    param1 = self._ctx.convert.primitive_class_instances[int]
    param2 = self._ctx.convert.primitive_class_instances[str]
    param_var = param1.to_variable(self._ctx.root_node)
    str_binding = param_var.AddBinding(param2, [], self._ctx.root_node)
    instance = self._ctx.convert.tuple_to_value((param_var,))
    view = {param_var: str_binding}
    pytd_type = instance.to_type(self._ctx.root_node, seen=None, view=view)
    self.assertEqual(pytd_type.parameters[0],
                     pytd.NamedType("builtins.str"))

  def test_to_type_with_view_and_empty_param(self):
    instance = abstract.List([], self._ctx)
    pytd_type = instance.to_type(self._ctx.root_node, seen=None, view={})
    self.assertEqual("builtins.list", pytd_type.name)
    self.assertSequenceEqual((pytd.NothingType(),), pytd_type.parameters)

  def test_typing_container(self):
    cls = self._ctx.convert.list_type
    container = abstract.AnnotationContainer("List", self._ctx, cls)
    expected = pytd.GenericType(pytd.NamedType("builtins.list"),
                                (pytd.AnythingType(),))
    actual = container.get_instance_type(self._ctx.root_node)
    self.assertEqual(expected, actual)


# TODO(rechen): Test InterpreterFunction.
class FunctionTest(AbstractTestBase):

  def _make_pytd_function(self, params, name="f"):
    pytd_params = []
    for i, p in enumerate(params):
      p_type = pytd.ClassType(p.name)
      p_type.cls = p
      pytd_params.append(
          pytd.Parameter(function.argname(i), p_type, False, False, None))
    pytd_sig = pytd.Signature(
        tuple(pytd_params), None, None, pytd.AnythingType(), (), ())
    sig = function.PyTDSignature(name, pytd_sig, self._ctx)
    return abstract.PyTDFunction(name, (sig,), pytd.MethodTypes.METHOD,
                                 self._ctx)

  def _call_pytd_function(self, f, args):
    b = f.to_binding(self._ctx.root_node)
    return f.call(self._ctx.root_node, b, function.Args(posargs=args))

  def test_call_with_empty_arg(self):
    self.assertRaises(AssertionError, self._call_pytd_function,
                      self._make_pytd_function(params=()),
                      (self._ctx.program.NewVariable(),))

  def test_call_with_bad_arg(self):
    f = self._make_pytd_function(
        (self._ctx.loader.lookup_builtin("builtins.str"),))
    arg = self._ctx.convert.primitive_class_instances[int].to_variable(
        self._ctx.root_node)
    self.assertRaises(
        function.WrongArgTypes, self._call_pytd_function, f, (arg,))

  def test_simple_call(self):
    f = self._make_pytd_function(
        (self._ctx.loader.lookup_builtin("builtins.str"),))
    arg = self._ctx.convert.primitive_class_instances[str].to_variable(
        self._ctx.root_node)
    node, ret = self._call_pytd_function(f, (arg,))
    self.assertIs(node, self._ctx.root_node)
    retval, = ret.bindings
    self.assertIs(retval.data, self._ctx.convert.unsolvable)

  def test_call_with_multiple_arg_bindings(self):
    f = self._make_pytd_function(
        (self._ctx.loader.lookup_builtin("builtins.str"),))
    arg = self._ctx.program.NewVariable()
    arg.AddBinding(self._ctx.convert.primitive_class_instances[str], [],
                   self._ctx.root_node)
    arg.AddBinding(self._ctx.convert.primitive_class_instances[int], [],
                   self._ctx.root_node)
    node, ret = self._call_pytd_function(f, (arg,))
    self.assertIs(node, self._ctx.root_node)
    retval, = ret.bindings
    self.assertIs(retval.data, self._ctx.convert.unsolvable)

  def test_call_with_skipped_combination(self):
    f = self._make_pytd_function(
        (self._ctx.loader.lookup_builtin("builtins.str"),))
    node = self._ctx.root_node.ConnectNew()
    arg = self._ctx.convert.primitive_class_instances[str].to_variable(node)
    node, ret = self._call_pytd_function(f, (arg,))
    self.assertIs(node, self._ctx.root_node)
    self.assertFalse(ret.bindings)

  def test_signature_from_pytd(self):
    # def f(self: Any, *args: Any)
    self_param = pytd.Parameter("self", pytd.AnythingType(), False, False, None)
    args_param = pytd.Parameter("args", pytd.AnythingType(), False, True, None)
    sig = function.Signature.from_pytd(
        self._ctx, "f",
        pytd.Signature((self_param,), args_param, None, pytd.AnythingType(), (),
                       ()))
    self.assertEqual(repr(sig), "def f(self: Any, *args: Any) -> Any")
    self.assertEqual(sig.name, "f")
    self.assertSequenceEqual(sig.param_names, ("self",))
    self.assertEqual(sig.varargs_name, "args")
    self.assertFalse(sig.kwonly_params)
    self.assertIs(sig.kwargs_name, None)
    self.assertSetEqual(set(sig.annotations), {"self", "args", "return"})
    self.assertTrue(sig.has_return_annotation)
    self.assertTrue(sig.has_param_annotations)

  def test_signature_from_callable(self):
    # Callable[[int, str], Any]
    params = {0: self._ctx.convert.int_type, 1: self._ctx.convert.str_type}
    params[abstract_utils.ARGS] = abstract.Union((params[0], params[1]),
                                                 self._ctx)
    params[abstract_utils.RET] = self._ctx.convert.unsolvable
    callable_val = abstract.CallableClass(self._ctx.convert.function_type,
                                          params, self._ctx)
    sig = function.Signature.from_callable(callable_val)
    self.assertEqual(repr(sig), "def <callable>(_0: int, _1: str) -> Any")
    self.assertEqual(sig.name, "<callable>")
    self.assertSequenceEqual(sig.param_names, ("_0", "_1"))
    self.assertIs(sig.varargs_name, None)
    self.assertFalse(sig.kwonly_params)
    self.assertIs(sig.kwargs_name, None)
    self.assertCountEqual(sig.annotations.keys(), sig.param_names)
    self.assertFalse(sig.has_return_annotation)
    self.assertTrue(sig.has_param_annotations)

  def test_signature_annotations(self):
    # def f(self: Any, *args: Any)
    self_param = pytd.Parameter("self", pytd.AnythingType(), False, False, None)
    # Imitate the parser's conversion of '*args: Any' to '*args: Tuple[Any]'.
    tup = pytd.ClassType("builtins.tuple")
    tup.cls = self._ctx.convert.tuple_type.pytd_cls
    any_tuple = pytd.GenericType(tup, (pytd.AnythingType(),))
    args_param = pytd.Parameter("args", any_tuple, False, True, None)
    sig = function.Signature.from_pytd(
        self._ctx, "f",
        pytd.Signature((self_param,), args_param, None, pytd.AnythingType(), (),
                       ()))
    self.assertEqual(repr(sig),
                     "def f(self: Any, *args: Tuple[Any, ...]) -> Any")
    self.assertIs(sig.annotations["self"], self._ctx.convert.unsolvable)
    args_type = sig.annotations["args"]
    self.assertIsInstance(args_type, abstract.ParameterizedClass)
    self.assertIs(args_type.base_cls, self._ctx.convert.tuple_type)
    self.assertListEqual(
        list(args_type.formal_type_parameters.items()),
        [(abstract_utils.T, self._ctx.convert.unsolvable)])
    self.assertIs(sig.drop_first_parameter().annotations["args"], args_type)

  def test_signature_annotations_existence(self):
    # def f(v: "X") -> "Y"
    sig = function.Signature(
        name="f",
        param_names=("v",),
        varargs_name=None,
        kwonly_params=(),
        kwargs_name=None,
        defaults={},
        annotations={},
    )
    self.assertFalse(sig.has_param_annotations)
    self.assertFalse(sig.has_return_annotation)
    sig.set_annotation("v", self._ctx.convert.unsolvable)
    self.assertTrue(sig.has_param_annotations)
    self.assertFalse(sig.has_return_annotation)
    sig.set_annotation("return", self._ctx.convert.unsolvable)
    self.assertTrue(sig.has_param_annotations)
    self.assertTrue(sig.has_return_annotation)

  def test_signature_posarg_only_param_count(self):
    # def f(x): ...
    sig = function.Signature(
        name="f",
        param_names=("x",),
        varargs_name=None,
        kwonly_params=(),
        kwargs_name=None,
        defaults={},
        annotations={},
    )
    self.assertEqual(repr(sig), "def f(x) -> Any")
    self.assertEqual(sig.mandatory_param_count(), 1)
    self.assertEqual(sig.maximum_param_count(), 1)

  def test_signature_posarg_and_kwarg_param_count(self):
    # def f(x, y=None): ...
    sig = function.Signature(
        name="f",
        param_names=(
            "x",
            "y",
        ),
        varargs_name=None,
        kwonly_params=(),
        kwargs_name=None,
        defaults={"y": self._ctx.convert.none_type.to_variable(self._node)},
        annotations={},
    )
    self.assertEqual(repr(sig), "def f(x, y = None) -> Any")
    self.assertEqual(sig.mandatory_param_count(), 1)
    self.assertEqual(sig.maximum_param_count(), 2)

  def test_signature_varargs_param_count(self):
    # def f(*args): ...
    sig = function.Signature(
        name="f",
        param_names=(),
        varargs_name="args",
        kwonly_params=(),
        kwargs_name=None,
        defaults={},
        annotations={},
    )
    self.assertEqual(repr(sig), "def f(*args) -> Any")
    self.assertEqual(sig.mandatory_param_count(), 0)
    self.assertIsNone(sig.maximum_param_count())

  def test_signature_kwargs_param_count(self):
    # def f(**kwargs): ...
    sig = function.Signature(
        name="f",
        param_names=(),
        varargs_name=None,
        kwonly_params=(),
        kwargs_name="kwargs",
        defaults={},
        annotations={},
    )
    self.assertEqual(repr(sig), "def f(**kwargs) -> Any")
    self.assertEqual(sig.mandatory_param_count(), 0)
    self.assertIsNone(sig.maximum_param_count())

  def test_signature_kwonly_param_count(self):
    # def f(*, y=None): ...
    sig = function.Signature(
        name="f",
        param_names=(),
        varargs_name=None,
        kwonly_params=("y",),
        kwargs_name=None,
        defaults={"y": self._ctx.convert.none_type.to_variable(self._node)},
        annotations={},
    )
    self.assertEqual(repr(sig), "def f(*, y = None) -> Any")
    self.assertEqual(sig.mandatory_param_count(), 0)
    self.assertEqual(sig.maximum_param_count(), 1)

  def test_signature_has_param(self):
    # def f(x, *args, y, **kwargs): ...
    sig = function.Signature(
        name="f",
        param_names=("x",),
        varargs_name="args",
        kwonly_params={"y"},
        kwargs_name="kwargs",
        defaults={},
        annotations={},
    )
    self.assertEqual(repr(sig), "def f(x, *args, y, **kwargs) -> Any")
    for param in ("x", "args", "y", "kwargs"):
      self.assertTrue(sig.has_param(param))
    self.assertFalse(sig.has_param("rumpelstiltskin"))

  def test_signature_insert_varargs_and_kwargs(self):
    # def f(x, *args, y, **kwargs): ...
    sig = function.Signature(
        name="f",
        param_names=("x",),
        varargs_name="args",
        kwonly_params={"y"},
        kwargs_name="kwargs",
        defaults={},
        annotations={},
    )
    # f(1, 2, y=3, z=4)
    int_inst = self._ctx.convert.primitive_class_instances[int]
    int_binding = int_inst.to_binding(self._node)
    arg_dict = {
        "x": int_binding, "_1": int_binding, "y": int_binding, "z": int_binding}
    sig = sig.insert_varargs_and_kwargs(arg_dict)
    self.assertEqual(sig.name, "f")
    self.assertSequenceEqual(sig.param_names, ("x", "_1", "z"))
    self.assertEqual(sig.varargs_name, "args")
    self.assertSetEqual(sig.kwonly_params, {"y"})
    self.assertEqual(sig.kwargs_name, "kwargs")
    self.assertFalse(sig.annotations)

  def test_signature_del_param_annotation(self):
    # def f(x) -> int: ...
    sig = function.Signature(
        name="f",
        param_names=("x",),
        varargs_name=None,
        kwonly_params=(),
        kwargs_name=None,
        defaults={},
        annotations={
            "x": self._ctx.convert.unsolvable,
            "return": self._ctx.convert.unsolvable
        },
    )
    sig.del_annotation("x")
    self.assertCountEqual(sig.annotations.keys(), {"return"})
    self.assertFalse(sig.has_param_annotations)
    self.assertTrue(sig.has_return_annotation)

  def test_signature_del_return_annotation(self):
    # def f(x) -> int: ...
    sig = function.Signature(
        name="f",
        param_names=("x",),
        varargs_name=None,
        kwonly_params=(),
        kwargs_name=None,
        defaults={},
        annotations={
            "x": self._ctx.convert.unsolvable,
            "return": self._ctx.convert.unsolvable
        },
    )
    sig.del_annotation("return")
    self.assertCountEqual(sig.annotations.keys(), {"x"})
    self.assertTrue(sig.has_param_annotations)
    self.assertFalse(sig.has_return_annotation)

  def test_signature_del_nonexistent_annotation(self):
    # def f(): ...
    sig = function.Signature(
        name="f",
        param_names=(),
        varargs_name=None,
        kwonly_params=(),
        kwargs_name=None,
        defaults={},
        annotations={},
    )
    self.assertRaises(KeyError, sig.del_annotation, "rumpelstiltskin")

  def test_constructor_args(self):
    f = abstract.PyTDFunction.make("open", self._ctx, "builtins")
    self.assertEqual(f.full_name, "builtins.open")
    self.assertCountEqual(
        {sig.pytd_sig for sig in f.signatures},
        self._ctx.loader.lookup_builtin("builtins.open").signatures)
    self.assertIs(f.kind, pytd.MethodTypes.METHOD)
    self.assertIs(f.ctx.vm, self._ctx.vm)

  def test_constructor_args_pyval(self):
    sig = pytd.Signature((), None, None, pytd.AnythingType(), (), ())
    pyval = pytd.Function("blah", (sig,), pytd.MethodTypes.STATICMETHOD, 0)
    f = abstract.PyTDFunction.make("open", self._ctx, "builtins", pyval=pyval)
    self.assertEqual(f.full_name, "builtins.open")
    f_sig, = f.signatures
    self.assertIs(f_sig.pytd_sig, sig)
    self.assertIs(f.kind, pytd.MethodTypes.STATICMETHOD)
    self.assertIs(f.ctx.vm, self._ctx.vm)

  def test_get_constructor_args(self):
    f = abstract.PyTDFunction.make(
        "TypeVar", self._ctx, "typing", pyval_name="_typevar_new")
    self.assertEqual(f.full_name, "typing.TypeVar")
    self.assertCountEqual({sig.pytd_sig for sig in f.signatures},
                          self._ctx.loader.import_name("typing").Lookup(
                              "typing._typevar_new").signatures)
    self.assertIs(f.kind, pytd.MethodTypes.METHOD)
    self.assertIs(f.ctx.vm, self._ctx.vm)

  def test_bound_function_repr(self):
    f = self._make_pytd_function(params=())
    callself = self._ctx.program.NewVariable(
        [abstract.BaseValue(name, self._ctx) for name in ("test1", "test2")],
        [], self._ctx.root_node)
    bound = abstract.BoundFunction(callself, f)
    self.assertCountEqual(bound.repr_names(), ["test1.f", "test2.f"])
    self.assertRegex(repr(bound), r"test(1|2)\.f")

  def test_bound_function_callself_repr(self):
    f = self._make_pytd_function(params=())
    callself = self._ctx.program.NewVariable(
        [abstract.BaseValue("test", self._ctx)], [], self._ctx.root_node)
    bound = abstract.BoundFunction(callself, f)
    callself_repr = lambda v: v.name + "foo"
    self.assertCountEqual(bound.repr_names(callself_repr), ["testfoo.f"])

  def test_bound_function_nested_repr(self):
    f = self._make_pytd_function(params=())
    callself1 = self._ctx.program.NewVariable(
        [abstract.BaseValue("test1", self._ctx)], [], self._ctx.root_node)
    bound1 = abstract.BoundFunction(callself1, f)
    callself2 = self._ctx.program.NewVariable(
        [abstract.BaseValue("test2", self._ctx)], [], self._ctx.root_node)
    bound2 = abstract.BoundFunction(callself2, bound1)
    # `bound2` is BoundFunction(test2, BoundFunction(test1, f))
    self.assertCountEqual(bound2.repr_names(), ["test2.f"])

  def test_bound_function_repr_no_callself(self):
    f = self._make_pytd_function(params=())
    callself = self._ctx.program.NewVariable()
    bound = abstract.BoundFunction(callself, f)
    self.assertCountEqual(bound.repr_names(), ["<class>.f"])

  def test_bound_function_repr_replace_parent(self):
    f = self._make_pytd_function(params=(), name="foo.f")
    callself = self._ctx.program.NewVariable(
        [abstract.BaseValue("test", self._ctx)], [], self._ctx.root_node)
    bound = abstract.BoundFunction(callself, f)
    self.assertCountEqual(bound.repr_names(), ["test.f"])


class AbstractMethodsTest(AbstractTestBase):

  def test_abstract_method(self):
    func = abstract.Function("f", self._ctx).to_variable(self._ctx.root_node)
    func.data[0].is_abstract = True
    cls = abstract.InterpreterClass("X", [], {"f": func}, None, self._ctx)
    self.assertCountEqual(cls.abstract_methods, {"f"})

  def test_inherited_abstract_method(self):
    sized_pytd = self._ctx.loader.typing.Lookup("typing.Sized")
    sized = self._ctx.convert.constant_to_value(sized_pytd, {},
                                                self._ctx.root_node)
    cls = abstract.InterpreterClass("X",
                                    [sized.to_variable(self._ctx.root_node)],
                                    {}, None, self._ctx)
    self.assertCountEqual(cls.abstract_methods, {"__len__"})

  def test_overridden_abstract_method(self):
    sized_pytd = self._ctx.loader.typing.Lookup("typing.Sized")
    sized = self._ctx.convert.constant_to_value(sized_pytd, {},
                                                self._ctx.root_node)
    bases = [sized.to_variable(self._ctx.root_node)]
    members = {"__len__": self._ctx.new_unsolvable(self._ctx.root_node)}
    cls = abstract.InterpreterClass("X", bases, members, None, self._ctx)
    self.assertFalse(cls.abstract_methods)

  def test_overridden_abstract_method_still_abstract(self):
    sized_pytd = self._ctx.loader.typing.Lookup("typing.Sized")
    sized = self._ctx.convert.constant_to_value(sized_pytd, {},
                                                self._ctx.root_node)
    bases = [sized.to_variable(self._ctx.root_node)]
    func = abstract.Function("__len__", self._ctx)
    func.is_abstract = True
    members = {"__len__": func.to_variable(self._ctx.root_node)}
    cls = abstract.InterpreterClass("X", bases, members, None, self._ctx)
    self.assertCountEqual(cls.abstract_methods, {"__len__"})


class SimpleFunctionTest(AbstractTestBase):

  def _make_func(self, name="_", param_names=None, varargs_name=None,
                 kwonly_params=(), kwargs_name=None, defaults=(),
                 annotations=None):
    return abstract.SimpleFunction(name, param_names or (), varargs_name,
                                   kwonly_params, kwargs_name, defaults,
                                   annotations or {}, self._ctx)

  def _simple_sig(self, param_types, ret_type=None):
    annots = {("_%d" % i): t for i, t in enumerate(param_types)}
    params = tuple(annots.keys())
    if ret_type:
      annots["return"] = ret_type
    return self._make_func(param_names=params, annotations=annots)

  def test_simple_call(self):
    f = self._simple_sig([self._ctx.convert.str_type],
                         ret_type=self._ctx.convert.int_type)
    args = function.Args(
        (self._ctx.convert.build_string(self._ctx.root_node, "hello"),))
    node, ret = f.call(self._ctx.root_node, f, args)
    self.assertIs(node, self._ctx.root_node)
    ret_val, = ret.data
    self.assertEqual(ret_val.cls, self._ctx.convert.int_type)

  def test_call_with_bad_arg(self):
    f = self._make_func(
        param_names=("test",), annotations={"test": self._ctx.convert.str_type})
    args = function.Args((self._ctx.convert.build_int(self._ctx.root_node),))
    self.assertRaises(function.WrongArgTypes, f.call, self._ctx.root_node, f,
                      args)

  def test_call_with_no_args(self):
    f = self._simple_sig(
        [self._ctx.convert.str_type, self._ctx.convert.int_type])
    args = function.Args(())
    self.assertRaises(function.MissingParameter, f.call, self._ctx.root_node, f,
                      args)

  def test_call_with_multiple_arg_bindings(self):
    f = self._simple_sig([self._ctx.convert.str_type])
    arg = self._ctx.program.NewVariable()
    arg.AddBinding(self._ctx.convert.primitive_class_instances[str], [],
                   self._ctx.root_node)
    arg.AddBinding(self._ctx.convert.primitive_class_instances[int], [],
                   self._ctx.root_node)
    args = function.Args((arg,))
    node, ret = f.call(self._ctx.root_node, f, args)
    self.assertIs(node, self._ctx.root_node)
    self.assertIs(ret.data[0], self._ctx.convert.none)

  def test_call_with_varargs(self):
    f = self._make_func(
        varargs_name="arg",
        annotations={
            "arg": self._ctx.convert.str_type,
            "return": self._ctx.convert.str_type
        })
    starargs = self._ctx.convert.build_tuple(
        self._ctx.root_node,
        (self._ctx.convert.build_string(self._ctx.root_node, ""),))
    args = function.Args(posargs=(), starargs=starargs)
    node, ret = f.call(self._ctx.root_node, f, args)
    self.assertIs(node, self._ctx.root_node)
    self.assertIs(ret.data[0].cls, self._ctx.convert.str_type)

  def test_call_with_bad_varargs(self):
    f = self._make_func(
        varargs_name="arg", annotations={"arg": self._ctx.convert.str_type})
    starargs = self._ctx.convert.build_tuple(
        self._ctx.root_node,
        (self._ctx.convert.build_string(self._ctx.root_node, ""),
         self._ctx.convert.build_int(self._ctx.root_node)))
    args = function.Args(posargs=(), starargs=starargs)
    self.assertRaises(function.WrongArgTypes, f.call, self._ctx.root_node, f,
                      args)

  def test_call_with_multiple_varargs_bindings(self):
    f = self._make_func(
        varargs_name="arg", annotations={"arg": self._ctx.convert.str_type})
    arg = self._ctx.program.NewVariable()
    arg.AddBinding(self._ctx.convert.primitive_class_instances[str], [],
                   self._ctx.root_node)
    arg.AddBinding(self._ctx.convert.primitive_class_instances[int], [],
                   self._ctx.root_node)
    starargs = self._ctx.convert.build_tuple(self._ctx.root_node, (arg,))
    args = function.Args(posargs=(), starargs=starargs)
    f.call(self._ctx.root_node, f, args)

  def test_call_with_kwargs(self):
    f = self._make_func(
        kwargs_name="kwarg", annotations={"kwarg": self._ctx.convert.str_type})
    kwargs = abstract.Dict(self._ctx)
    kwargs.update(
        self._ctx.root_node, {
            "_1": self._ctx.convert.build_string(self._ctx.root_node, "1"),
            "_2": self._ctx.convert.build_string(self._ctx.root_node, "2")
        })
    kwargs = kwargs.to_variable(self._ctx.root_node)
    args = function.Args(
        posargs=(), namedargs=abstract.Dict(self._ctx), starstarargs=kwargs)
    f.call(self._ctx.root_node, f, args)

  def test_call_with_bad_kwargs(self):
    f = self._make_func(
        kwargs_name="kwarg", annotations={"kwarg": self._ctx.convert.str_type})
    kwargs = abstract.Dict(self._ctx)
    kwargs.update(self._ctx.root_node,
                  {"_1": self._ctx.convert.build_int(self._ctx.root_node)})
    kwargs = kwargs.to_variable(self._ctx.root_node)
    args = function.Args(
        posargs=(), namedargs=abstract.Dict(self._ctx), starstarargs=kwargs)
    self.assertRaises(function.WrongArgTypes, f.call, self._ctx.root_node, f,
                      args)

  def test_call_with_kwonly_args(self):
    f = self._make_func(
        param_names=("test",),
        kwonly_params=("a", "b"),
        annotations={
            "test": self._ctx.convert.str_type,
            "a": self._ctx.convert.str_type,
            "b": self._ctx.convert.str_type
        })
    kwargs = abstract.Dict(self._ctx)
    kwargs.update(
        self._ctx.root_node, {
            "a": self._ctx.convert.build_string(self._ctx.root_node, "2"),
            "b": self._ctx.convert.build_string(self._ctx.root_node, "3")
        })
    kwargs = kwargs.to_variable(self._ctx.root_node)
    args = function.Args(
        posargs=(self._ctx.convert.build_string(self._ctx.root_node, "1"),),
        namedargs=abstract.Dict(self._ctx),
        starstarargs=kwargs)
    f.call(self._ctx.root_node, f, args)
    kwargs = abstract.Dict(self._ctx)
    kwargs.update(
        self._ctx.root_node,
        {"b": self._ctx.convert.build_string(self._ctx.root_node, "3")})
    kwargs = kwargs.to_variable(self._ctx.root_node)
    args = function.Args(
        posargs=(self._ctx.convert.build_string(self._ctx.root_node, "1"),
                 self._ctx.convert.build_int(self._ctx.root_node)),
        namedargs=abstract.Dict(self._ctx),
        starstarargs=kwargs)
    self.assertRaises(function.MissingParameter, f.call, self._ctx.root_node, f,
                      args)

  def test_call_with_all_args(self):
    f = self._make_func(
        param_names=("a", "b", "c"),
        varargs_name="arg",
        kwargs_name="kwarg",
        defaults=(self._ctx.convert.build_int(self._ctx.root_node),),
        annotations={
            "a": self._ctx.convert.str_type,
            "b": self._ctx.convert.int_type,
            "c": self._ctx.convert.int_type,
            "arg": self._ctx.convert.primitive_classes[float],
            "kwarg": self._ctx.convert.primitive_classes[bool]
        })
    posargs = (self._ctx.convert.build_string(self._ctx.root_node, "1"),
               self._ctx.convert.build_int(self._ctx.root_node))
    float_inst = self._ctx.convert.primitive_class_instances[float]
    stararg = self._ctx.convert.build_tuple(
        self._ctx.root_node, (float_inst.to_variable(self._ctx.root_node),))
    namedargs = abstract.Dict(self._ctx)
    kwarg = abstract.Dict(self._ctx)
    kwarg.update(
        self._ctx.root_node, {
            "x": self._ctx.convert.build_bool(self._ctx.root_node),
            "y": self._ctx.convert.build_bool(self._ctx.root_node)
        })
    kwarg = kwarg.to_variable(self._ctx.root_node)
    args = function.Args(posargs, namedargs, stararg, kwarg)
    f.call(self._ctx.root_node, f, args)

  def test_call_with_defaults(self):
    f = self._make_func(
        param_names=("a", "b", "c"),
        defaults=(self._ctx.convert.build_int(self._ctx.root_node),),
        annotations={
            "a": self._ctx.convert.int_type,
            "b": self._ctx.convert.int_type,
            "c": self._ctx.convert.int_type
        })
    args = function.Args(
        posargs=(self._ctx.convert.build_int(self._ctx.root_node),
                 self._ctx.convert.build_int(self._ctx.root_node)))
    f.call(self._ctx.root_node, f, args)
    args = function.Args(
        posargs=(self._ctx.convert.build_int(self._ctx.root_node),
                 self._ctx.convert.build_int(self._ctx.root_node),
                 self._ctx.convert.build_int(self._ctx.root_node)))
    f.call(self._ctx.root_node, f, args)
    args = function.Args(
        posargs=(self._ctx.convert.build_int(self._ctx.root_node),))
    self.assertRaises(function.MissingParameter, f.call, self._ctx.root_node, f,
                      args)

  def test_call_with_bad_default(self):
    f = self._make_func(
        param_names=("a", "b"),
        defaults=(self._ctx.convert.build_string(self._ctx.root_node, ""),),
        annotations={
            "a": self._ctx.convert.int_type,
            "b": self._ctx.convert.str_type
        })
    args = function.Args(
        posargs=(self._ctx.convert.build_int(self._ctx.root_node),
                 self._ctx.convert.build_int(self._ctx.root_node)))
    self.assertRaises(function.WrongArgTypes, f.call, self._ctx.root_node, f,
                      args)

  def test_call_with_duplicate_keyword(self):
    f = self._simple_sig([self._ctx.convert.int_type] * 2)
    args = function.Args(
        posargs=(self._ctx.convert.build_int(self._ctx.root_node),
                 self._ctx.convert.build_int(self._ctx.root_node)),
        namedargs={"_1": self._ctx.convert.build_int(self._ctx.root_node)})
    self.assertRaises(function.DuplicateKeyword, f.call, self._ctx.root_node, f,
                      args)

  def test_call_with_wrong_arg_count(self):
    f = self._simple_sig([self._ctx.convert.int_type])
    args = function.Args(
        posargs=(self._ctx.convert.build_int(self._ctx.root_node),
                 self._ctx.convert.build_int(self._ctx.root_node)))
    self.assertRaises(function.WrongArgCount, f.call, self._ctx.root_node, f,
                      args)

  def test_change_defaults(self):
    f = self._make_func(
        param_names=("a", "b", "c"),
        defaults=(self._ctx.convert.build_int(self._ctx.root_node),))
    args = function.Args(
        posargs=(self._ctx.convert.build_int(self._ctx.root_node),
                 self._ctx.convert.build_int(self._ctx.root_node)))
    f.call(self._ctx.root_node, f, args)
    new_defaults = self._ctx.convert.build_tuple(
        self._ctx.root_node,
        (self._ctx.convert.build_int(self._ctx.root_node),
         self._ctx.convert.build_int(self._ctx.root_node)))
    f.set_function_defaults(self._ctx.root_node, new_defaults)
    f.call(self._ctx.root_node, f, args)
    args = function.Args(
        posargs=(self._ctx.convert.build_int(self._ctx.root_node),))
    f.call(self._ctx.root_node, f, args)

  def test_call_with_type_parameter(self):
    ret_cls = abstract.ParameterizedClass(
        self._ctx.convert.list_type,
        {abstract_utils.T: abstract.TypeParameter(abstract_utils.T, self._ctx)},
        self._ctx)
    f = self._make_func(
        param_names=("test",),
        annotations={
            "test": abstract.TypeParameter(abstract_utils.T, self._ctx),
            "return": ret_cls
        })
    args = function.Args(
        posargs=(self._ctx.convert.build_int(self._ctx.root_node),))
    _, ret = f.call(self._ctx.root_node, f, args)
    # ret is an Instance(ParameterizedClass(list, {abstract_utils.T: int}))
    # but we really only care about T.
    self.assertIs(ret.data[0].cls.formal_type_parameters[abstract_utils.T],
                  self._ctx.convert.int_type)

  def test_signature_func_output_basic(self):
    node = self._ctx.root_node
    f = self._make_func(name="basic", param_names=("a", "b"))
    fp = self._ctx.pytd_convert.value_to_pytd_def(node, f, f.name)
    self.assertEqual(pytd_utils.Print(fp), "def basic(a, b) -> None: ...")

  def test_signature_func_output_annotations(self):
    node = self._ctx.root_node
    f = self._make_func(
        name="annots",
        param_names=("a", "b"),
        annotations={
            "a": self._ctx.convert.int_type,
            "b": self._ctx.convert.str_type,
            "return": self._ctx.convert.int_type
        })
    fp = self._ctx.pytd_convert.value_to_pytd_def(node, f, f.name)
    self.assertEqual(pytd_utils.Print(fp),
                     "def annots(a: int, b: str) -> int: ...")

  def test_signature_func_output(self):
    node = self._ctx.root_node
    dict_type = abstract.ParameterizedClass(
        self._ctx.convert.dict_type, {
            abstract_utils.K: self._ctx.convert.str_type,
            abstract_utils.V: self._ctx.convert.int_type
        }, self._ctx)
    f = self._make_func(
        name="test",
        param_names=("a", "b"),
        varargs_name="c",
        kwonly_params=("d", "e"),
        kwargs_name="f",
        defaults={
            "b": self._ctx.convert.build_int(node),
            "d": self._ctx.convert.build_int(node)
        },
        annotations={
            "a": self._ctx.convert.str_type,
            "b": self._ctx.convert.int_type,
            "c": self._ctx.convert.str_type,
            "d": dict_type,
            "e": self._ctx.convert.int_type,
            "f": self._ctx.convert.str_type,
            "return": self._ctx.convert.str_type
        })
    fp = self._ctx.pytd_convert.value_to_pytd_def(node, f, f.name)
    f_str = ("def test(a: str, b: int = ..., *c: str, d: Dict[str, int] = ...,"
             " e: int, **f: str) -> str: ...")
    self.assertEqual(pytd_utils.Print(fp), f_str)


class AbstractTest(AbstractTestBase):

  def test_interpreter_class_official_name(self):
    cls = abstract.InterpreterClass("X", [], {}, None, self._ctx)
    cls.update_official_name("Z")
    self.assertEqual(cls.official_name, "Z")
    cls.update_official_name("A")  # takes effect because A < Z
    self.assertEqual(cls.official_name, "A")
    cls.update_official_name("Z")  # no effect
    self.assertEqual(cls.official_name, "A")
    cls.update_official_name("X")  # takes effect because X == cls.name
    self.assertEqual(cls.official_name, "X")
    cls.update_official_name("A")  # no effect
    self.assertEqual(cls.official_name, "X")

  def test_type_parameter_official_name(self):
    param = abstract.TypeParameter("T", self._ctx)
    self._ctx.vm.frame = frame_state.SimpleFrame()  # for error logging
    param.update_official_name("T")
    self.assertFalse(self._ctx.errorlog.has_error())
    param.update_official_name("Q")
    self.assertTrue(self._ctx.errorlog.has_error())

  def test_type_parameter_equality(self):
    param1 = abstract.TypeParameter("S", self._ctx)
    param2 = abstract.TypeParameter("T", self._ctx)
    cls = abstract.InterpreterClass("S", [], {}, None, self._ctx)
    self.assertEqual(param1, param1)
    self.assertNotEqual(param1, param2)
    self.assertNotEqual(param1, cls)

  def test_union_equality(self):
    union1 = abstract.Union((self._ctx.convert.unsolvable,), self._ctx)
    union2 = abstract.Union((self._ctx.convert.none,), self._ctx)
    cls = abstract.InterpreterClass("Union", [], {}, None, self._ctx)
    self.assertEqual(union1, union1)
    self.assertNotEqual(union1, union2)
    self.assertNotEqual(union1, cls)

  def test_instantiate_type_parameter_type(self):
    params = {
        abstract_utils.T: abstract.TypeParameter(abstract_utils.T, self._ctx)
    }
    cls = abstract.ParameterizedClass(self._ctx.convert.type_type, params,
                                      self._ctx)
    self.assertListEqual(
        cls.instantiate(self._node).data, [self._ctx.convert.unsolvable])

  def test_super_type(self):
    supercls = special_builtins.Super(self._ctx)
    self.assertEqual(supercls.cls, self._ctx.convert.type_type)

  def test_instantiate_interpreter_class(self):
    cls = abstract.InterpreterClass("X", [], {}, None, self._ctx)
    # When there is no current frame, create a new instance every time.
    v1 = abstract_utils.get_atomic_value(cls.instantiate(self._node))
    v2 = abstract_utils.get_atomic_value(cls.instantiate(self._node))
    self.assertIsNot(v1, v2)
    # Create one instance per opcode.
    fake_opcode = object()
    self._ctx.vm.push_frame(frame_state.SimpleFrame(fake_opcode))
    v3 = abstract_utils.get_atomic_value(cls.instantiate(self._node))
    v4 = abstract_utils.get_atomic_value(cls.instantiate(self._node))
    self.assertIsNot(v1, v3)
    self.assertIsNot(v2, v3)
    self.assertIs(v3, v4)

  def test_set_module_on_module(self):
    # A module's 'module' attribute should always remain None, and no one
    # should attempt to set it to something besides the module's name or None.
    ast = pytd_utils.CreateModule("some_mod")
    mod = abstract.Module(self._ctx, ast.name, {}, ast)
    mod.module = ast.name
    self.assertIsNone(mod.module)
    self.assertEqual(ast.name, mod.full_name)
    mod.module = None
    self.assertIsNone(mod.module)
    self.assertEqual(ast.name, mod.full_name)
    def set_module():
      mod.module = "other_mod"
    self.assertRaises(AssertionError, set_module)

  def test_call_type_parameter_instance(self):
    instance = abstract.Instance(self._ctx.convert.list_type, self._ctx)
    instance.merge_instance_type_parameter(
        self._ctx.root_node, abstract_utils.T,
        self._ctx.convert.int_type.to_variable(self._ctx.root_node))
    t = abstract.TypeParameter(abstract_utils.T, self._ctx)
    t_instance = abstract.TypeParameterInstance(t, instance, self._ctx)
    node, ret = t_instance.call(self._node, t_instance.to_binding(self._node),
                                function.Args(posargs=()))
    self.assertIs(node, self._node)
    retval, = ret.data
    self.assertEqual(retval.cls, self._ctx.convert.int_type)

  def test_call_empty_type_parameter_instance(self):
    instance = abstract.Instance(self._ctx.convert.list_type, self._ctx)
    t = abstract.TypeParameter(abstract_utils.T, self._ctx)
    t_instance = abstract.TypeParameterInstance(t, instance, self._ctx)
    node, ret = t_instance.call(self._node, t_instance.to_binding(self._node),
                                function.Args(posargs=()))
    self.assertIs(node, self._node)
    retval, = ret.data
    self.assertIs(retval, self._ctx.convert.empty)

  def test_call_type_parameter_instance_with_wrong_args(self):
    instance = abstract.Instance(self._ctx.convert.list_type, self._ctx)
    instance.merge_instance_type_parameter(
        self._ctx.root_node, abstract_utils.T,
        self._ctx.convert.int_type.to_variable(self._ctx.root_node))
    t = abstract.TypeParameter(abstract_utils.T, self._ctx)
    t_instance = abstract.TypeParameterInstance(t, instance, self._ctx)
    posargs = (self._ctx.new_unsolvable(self._node),) * 3
    node, ret = t_instance.call(self._node, t_instance.to_binding(self._node),
                                function.Args(posargs=posargs))
    self.assertIs(node, self._node)
    self.assertTrue(ret.bindings)
    error, = self._ctx.errorlog
    self.assertEqual(error.name, "wrong-arg-count")

  def test_instantiate_tuple_class_for_sub(self):
    type_param = abstract.TypeParameter(abstract_utils.K, self._ctx)
    cls = abstract.TupleClass(self._ctx.convert.tuple_type, {
        0: type_param,
        abstract_utils.T: type_param
    }, self._ctx)
    # Instantiate the tuple class.
    subst_value = cls.instantiate(self._ctx.root_node,
                                  abstract_utils.DUMMY_CONTAINER)
    # Recover the class from the instance.
    subbed_cls = self._ctx.annotation_utils.sub_one_annotation(
        self._ctx.root_node, type_param, [{
            abstract_utils.K: subst_value
        }])
    self.assertEqual(cls, subbed_cls)

  def test_singleton(self):
    self.assertIs(
        abstract.Unsolvable(self._ctx), abstract.Unsolvable(self._ctx))

  def test_singleton_subclass(self):
    self.assertIs(abstract.Empty(self._ctx), abstract.Empty(self._ctx))
    self.assertIsNot(abstract.Deleted(self._ctx), abstract.Empty(self._ctx))


if __name__ == "__main__":
  unittest.main()



import string
import sys
from typing import Any, Callable, Iterable, Mapping, Sequence, Text
from typing_extensions import SupportsIndex

from markupsafe._compat import text_type
from markupsafe._native import escape as escape, escape_silent as escape_silent, soft_unicode as soft_unicode

class Markup(text_type):
    def __new__(cls, base: Text = ..., encoding: Text | None = ..., errors: Text = ...) -> Markup: ...
    def __html__(self) -> Markup: ...
    def __add__(self, other: text_type) -> Markup: ...
    def __radd__(self, other: text_type) -> Markup: ...
    def __mul__(self, num: int) -> Markup: ...  # type: ignore[override]
    def __rmul__(self, num: int) -> Markup: ...  # type: ignore[override]
    def __mod__(self, *args: Any) -> Markup: ...
    def join(self, seq: Iterable[text_type]) -> Markup: ...
    def split(self, sep: text_type | None = ..., maxsplit: SupportsIndex = ...) -> list[Markup]: ...  # type: ignore[override]
    def rsplit(self, sep: text_type | None = ..., maxsplit: SupportsIndex = ...) -> list[Markup]: ...  # type: ignore[override]
    def splitlines(self, keepends: bool = ...) -> list[Markup]: ...  # type: ignore[override]
    def unescape(self) -> Text: ...
    def striptags(self) -> Text: ...
    @classmethod
    def escape(cls, s: text_type) -> Markup: ...  # noqa: F811
    def partition(self, sep: text_type) -> tuple[Markup, Markup, Markup]: ...
    def rpartition(self, sep: text_type) -> tuple[Markup, Markup, Markup]: ...
    def format(self, *args: Any, **kwargs: Any) -> Markup: ...
    def __html_format__(self, format_spec: text_type) -> Markup: ...
    def __getslice__(self, start: int, stop: int) -> Markup: ...
    def __getitem__(self, i: SupportsIndex | slice) -> Markup: ...
    def capitalize(self) -> Markup: ...
    def title(self) -> Markup: ...
    def lower(self) -> Markup: ...
    def upper(self) -> Markup: ...
    def swapcase(self) -> Markup: ...
    def replace(self, old: text_type, new: text_type, count: SupportsIndex = ...) -> Markup: ...
    def ljust(self, width: SupportsIndex, fillchar: text_type = ...) -> Markup: ...
    def rjust(self, width: SupportsIndex, fillchar: text_type = ...) -> Markup: ...
    def lstrip(self, chars: text_type | None = ...) -> Markup: ...
    def rstrip(self, chars: text_type | None = ...) -> Markup: ...
    def strip(self, chars: text_type | None = ...) -> Markup: ...
    def center(self, width: SupportsIndex, fillchar: text_type = ...) -> Markup: ...
    def zfill(self, width: SupportsIndex) -> Markup: ...
    def translate(self, table: Mapping[int, int | text_type | None] | Sequence[int | text_type | None]) -> Markup: ...
    if sys.version_info >= (3, 8):
        def expandtabs(self, tabsize: SupportsIndex = ...) -> Markup: ...
    else:
        def expandtabs(self, tabsize: int = ...) -> Markup: ...

class EscapeFormatter(string.Formatter):
    escape: Callable[[text_type], Markup]
    def __init__(self, escape: Callable[[text_type], Markup]) -> None: ...  # noqa: F811
    def format_field(self, value: text_type, format_spec: text_type) -> Markup: ...

if sys.version_info >= (3,):
    soft_str = soft_unicode



"""Utilities for quoting, name-mangling, etc."""

import re

from typing import List


# The partial prefix was originally `~` and some part of the protocol inference
# code relies on it sorting later than `typing.`, so we start the prefix with
# `z` to make sure it comes last in the sort order. To reproduce, remove the `z`
# and test tests/test_protocol_inference
PARTIAL = "z__pytype_partial__"
UNKNOWN = PARTIAL + "unknown"


def pack_partial(name: str) -> str:
  """Pack a name, for unpacking with unpack_partial()."""
  return PARTIAL + name.replace(".", PARTIAL)


def unpack_partial(name: str) -> str:
  """Convert e.g. "~int" to "int"."""
  assert isinstance(name, str)
  assert name.startswith(PARTIAL)
  return name[len(PARTIAL):].replace(PARTIAL, ".")


def is_partial(cls) -> bool:
  """Returns True if this is a partial class, e.g. "~list"."""
  if isinstance(cls, str):
    return cls.startswith(PARTIAL)
  elif hasattr(cls, "name"):
    return cls.name.startswith(PARTIAL)
  else:
    return False


def is_complete(cls) -> bool:
  return not is_partial(cls)


def unknown(idcode: int) -> str:
  return UNKNOWN + str(idcode)


def is_unknown(name: str) -> bool:
  return name.startswith(UNKNOWN)


def preprocess_pytd(text: str) -> str:
  """Replace ~ in a text pytd with PARTIAL."""
  return text.replace("~", PARTIAL)


def pack_namedtuple(name: str, fields: List[str]) -> str:
  """Generate a name for a namedtuple proxy class."""
  return "namedtuple_%s_%s" % (name, "_".join(fields))


def pack_namedtuple_base_class(name: str, index: int) -> str:
  """Generate a name for a namedtuple proxy base class."""
  return "namedtuple_%s_%d" % (name, index)


def unpack_namedtuple(name: str) -> str:
  """Retrieve the original namedtuple class name."""
  return re.sub(r"\bnamedtuple[-_]([^-_]+)[-_\w]*", r"\1", name)


def pack_newtype_base_class(name: str, index: int) -> str:
  """Generate a name for a NewType proxy base class."""
  return "newtype_%s_%d" % (name, index)



from logging import Logger
from typing import Pattern
from typing_extensions import TypedDict

class _FinalResultType(TypedDict):
    encoding: str
    confidence: float
    language: str

class _IntermediateResultType(TypedDict):
    encoding: str | None
    confidence: float
    language: str | None

class UniversalDetector(object):
    MINIMUM_THRESHOLD: float
    HIGH_BYTE_DETECTOR: Pattern[bytes]
    ESC_DETECTOR: Pattern[bytes]
    WIN_BYTE_DETECTOR: Pattern[bytes]
    ISO_WIN_MAP: dict[str, str]

    result: _IntermediateResultType
    done: bool
    lang_filter: int
    logger: Logger
    def __init__(self, lang_filter: int = ...) -> None: ...
    def reset(self) -> None: ...
    def feed(self, byte_str: bytes) -> None: ...
    def close(self) -> _FinalResultType: ...



import sys
from unittest.async_case import *

from .case import (
    FunctionTestCase as FunctionTestCase,
    SkipTest as SkipTest,
    TestCase as TestCase,
    expectedFailure as expectedFailure,
    skip as skip,
    skipIf as skipIf,
    skipUnless as skipUnless,
)

if sys.version_info >= (3, 8):
    from .case import addModuleCleanup as addModuleCleanup

from unittest.loader import *
from unittest.main import *
from unittest.result import TestResult as TestResult
from unittest.runner import *
from unittest.signals import *
from unittest.suite import *

def load_tests(loader: TestLoader, tests: TestSuite, pattern: str | None) -> TestSuite: ...



"""Config file processing."""

import collections
import logging
import os
import sys
import textwrap

from pytype import config as pytype_config
from pytype import file_utils
from pytype import utils
from pytype.tools import config


# A config item.
# Args:
#   default: the default value.
#   sample: a sample value.
#   arg_info: information about the corresponding command-line argument.
#   comment: help text.
Item = collections.namedtuple(
    'Item', ['default', 'sample', 'arg_info', 'comment'])


# Args:
#   flag: the name of the command-line flag.
#   to_command_line: a function to transform the value into a command-line arg.
#     If None, the value will be used directly. The result will be interpreted
#     as follows:
#     - If a bool, it will determine whether the flag is included.
#     - Elif it is empty, the flag will be omitted.
#     - Else, it will be considered the flag's command-line value.
ArgInfo = collections.namedtuple('ArgInfo', ['flag', 'to_command_line'])


# Generates both the default config and the sample config file. These items
# don't have ArgInfo populated, as it is needed only for pytype-single args.
ITEMS = {
    'exclude': Item(
        '', '**/*_test.py **/test_*.py', None,
        'Space-separated list of files or directories to exclude.'),
    'inputs': Item(
        '', '.', None,
        'Space-separated list of files or directories to process.'),
    'keep_going': Item(
        False, 'False', None,
        'Keep going past errors to analyze as many files as possible.'),
    'jobs': Item(
        1, '4', None,
        "Run N jobs in parallel. When 'auto' is used, this will be equivalent "
        'to the number of CPUs on the host system.'),
    'output': Item(
        '.pytype', '.pytype', None, 'All pytype output goes here.'),
    'pythonpath': Item(
        '', '.', None,
        'Paths to source code directories, separated by %r.' % os.pathsep),
    'python_version': Item(
        '', '{}.{}'.format(*sys.version_info[:2]),
        None, 'Python version (major.minor) of the target code.'),
}


REPORT_ERRORS_ITEMS = {
    'disable': Item(
        None, 'pyi-error', ArgInfo('--disable', ','.join),
        'Comma or space separated list of error names to ignore.'),
    'report_errors': Item(
        None, 'True', ArgInfo('--no-report-errors', lambda v: not v), None),
}


# The missing fields will be filled in by generate_sample_config_or_die.
def _pytype_single_items():
  """Args to pass through to pytype_single."""
  out = {}
  flags = pytype_config.FEATURE_FLAGS + pytype_config.EXPERIMENTAL_FLAGS
  for opt, _, default in flags:
    dest = opt.lstrip('-').replace('-', '_')
    default = str(default)
    out[dest] = Item(None, default, ArgInfo(opt, None), None)
  out.update(REPORT_ERRORS_ITEMS)
  return out


_PYTYPE_SINGLE_ITEMS = _pytype_single_items()


def get_pytype_single_item(name):
  # We want to avoid exposing this hard-coded list as much as possible so that
  # parser.pytype_single_args, which is guaranteed to match the actual args, is
  # used instead.
  return _PYTYPE_SINGLE_ITEMS[name]


def string_to_bool(s):
  return s == 'True' if s in ('True', 'False') else s


def concat_disabled_rules(s):
  return ','.join(t for t in s.split() if t)


def get_python_version(v):
  return v if v else utils.format_version(sys.version_info[:2])


def make_converters(cwd=None):
  """For items that need coaxing into their internal representations."""
  return {
      'exclude': lambda v: file_utils.expand_source_files(v, cwd),
      'keep_going': string_to_bool,
      'inputs': lambda v: file_utils.expand_source_files(v, cwd),
      'output': lambda v: file_utils.expand_path(v, cwd),
      'python_version': get_python_version,
      'pythonpath': lambda v: file_utils.expand_pythonpath(v, cwd),
      'disable': concat_disabled_rules,
  }


def _make_spaced_path_formatter(name):
  """Formatter for space-separated paths."""
  def format_spaced_path(p):
    out = []
    out.append('%s =' % name)
    out.extend('    %s' % entry for entry in p.split())
    return out
  return format_spaced_path


def _make_separated_path_formatter(name, sep):
  """Formatter for paths separated by a non-space token."""
  def format_separated_path(p):
    out = []
    out.append('%s =' % name)
    # Breaks the path after each instance of sep.
    for entry in p.replace(sep, sep + '\n').split('\n'):
      out.append('    %s' % entry)
    return out
  return format_separated_path


def make_formatters():
  return {
      'disable': _make_separated_path_formatter('disable', ','),
      'exclude': _make_spaced_path_formatter('exclude'),
      'inputs': _make_spaced_path_formatter('inputs'),
      'pythonpath': _make_separated_path_formatter('pythonpath', os.pathsep),
  }


def Config(*extra_variables):  # pylint: disable=invalid-name
  """Builds a Config class and returns an instance of it."""

  class Config:  # pylint: disable=redefined-outer-name
    """Configuration variables.

    A lightweight configuration class that reads in attributes from other
    objects and prettyprints itself. The intention is for each source of
    attributes (e.g., FileConfig) to do its own processing, then for Config to
    copy in the final results in the right order.
    """

    __slots__ = tuple(ITEMS) + extra_variables

    def populate_from(self, obj):
      """Populate self from another object's attributes."""
      for k in self.__slots__:
        if hasattr(obj, k):
          setattr(self, k, getattr(obj, k))

    def __str__(self):
      return '\n'.join(
          '%s = %r' % (k, getattr(self, k, None)) for k in self.__slots__)

  return Config()


class FileConfig:
  """Configuration variables from a file."""

  def read_from_file(self, filepath):
    """Read config from an INI-style file with a [pytype] section."""

    cfg = config.ConfigSection.create_from_file(filepath, 'pytype')
    if not cfg:
      return None
    converters = make_converters(cwd=os.path.dirname(filepath))
    for k, v in cfg.items():
      if k in converters:
        v = converters[k](v)
      setattr(self, k, v)
    return filepath


def generate_sample_config_or_die(filename, pytype_single_args):
  """Write out a sample config file."""

  if os.path.exists(filename):
    logging.critical('Not overwriting existing file: %s', filename)
    sys.exit(1)

  # Combine all arguments into one name -> Item dictionary.
  items = dict(ITEMS)
  assert set(_PYTYPE_SINGLE_ITEMS) == set(pytype_single_args)
  for key, item in _PYTYPE_SINGLE_ITEMS.items():
    if item.comment is None:
      items[key] = item._replace(default=pytype_single_args[key].default,
                                 comment=pytype_single_args[key].help)
    else:
      items[key] = item._replace(default=pytype_single_args[key].default)

  # Not using configparser's write method because it doesn't support comments.

  conf = [
      '# NOTE: All relative paths are relative to the location of this file.',
      '',
      '[pytype]',
      '',
  ]
  formatters = make_formatters()
  for key, item in items.items():
    conf.extend(textwrap.wrap(
        item.comment, 80, initial_indent='# ', subsequent_indent='# '))
    if key in formatters:
      conf.extend(formatters[key](item.sample))
    else:
      conf.append('%s = %s' % (key, item.sample))
    conf.append('')
  try:
    with open(filename, 'w') as f:
      f.write('\n'.join(conf))
  except IOError as e:
    logging.critical('Cannot write to %s:\n%s', filename, str(e))
    sys.exit(1)


def read_config_file_or_die(filepath):
  """Read config from filepath or from setup.cfg."""

  ret = FileConfig()
  if filepath:
    if not ret.read_from_file(filepath):
      logging.critical('Could not read config file: %s\n'
                       '  Generate a sample configuration via:\n'
                       '  pytype --generate-config sample.cfg', filepath)
      sys.exit(1)
  else:
    # Try reading from setup.cfg.
    filepath = config.find_config_file(os.getcwd())
    if filepath and ret.read_from_file(filepath):
      logging.info('Reading config from: %s', filepath)
    else:
      logging.info('No config file specified, and no [pytype] section in '
                   'setup.cfg. Using default configuration.')
  return ret



import sys
from typing import IO, Any, Iterable, NoReturn
from xml.sax.handler import ContentHandler, ErrorHandler
from xml.sax.xmlreader import Locator, XMLReader

class SAXException(Exception):
    def __init__(self, msg: str, exception: Exception | None = ...) -> None: ...
    def getMessage(self) -> str: ...
    def getException(self) -> Exception: ...
    def __getitem__(self, ix: Any) -> NoReturn: ...

class SAXParseException(SAXException):
    def __init__(self, msg: str, exception: Exception, locator: Locator) -> None: ...
    def getColumnNumber(self) -> int: ...
    def getLineNumber(self) -> int: ...
    def getPublicId(self): ...
    def getSystemId(self): ...

class SAXNotRecognizedException(SAXException): ...
class SAXNotSupportedException(SAXException): ...
class SAXReaderNotAvailable(SAXNotSupportedException): ...

default_parser_list: list[str]

if sys.version_info >= (3, 8):
    def make_parser(parser_list: Iterable[str] = ...) -> XMLReader: ...

else:
    def make_parser(parser_list: list[str] = ...) -> XMLReader: ...

def parse(source: str | IO[str] | IO[bytes], handler: ContentHandler, errorHandler: ErrorHandler = ...) -> None: ...
def parseString(string: bytes | str, handler: ContentHandler, errorHandler: ErrorHandler | None = ...) -> None: ...
def _create_parser(parser_name: str) -> XMLReader: ...



single_input: int
file_input: int
eval_input: int
decorator: int
decorators: int
decorated: int
async_funcdef: int
funcdef: int
parameters: int
typedargslist: int
tfpdef: int
varargslist: int
vfpdef: int
stmt: int
simple_stmt: int
small_stmt: int
expr_stmt: int
annassign: int
testlist_star_expr: int
augassign: int
del_stmt: int
pass_stmt: int
flow_stmt: int
break_stmt: int
continue_stmt: int
return_stmt: int
yield_stmt: int
raise_stmt: int
import_stmt: int
import_name: int
import_from: int
import_as_name: int
dotted_as_name: int
import_as_names: int
dotted_as_names: int
dotted_name: int
global_stmt: int
nonlocal_stmt: int
assert_stmt: int
compound_stmt: int
async_stmt: int
if_stmt: int
while_stmt: int
for_stmt: int
try_stmt: int
with_stmt: int
with_item: int
except_clause: int
suite: int
test: int
test_nocond: int
lambdef: int
lambdef_nocond: int
or_test: int
and_test: int
not_test: int
comparison: int
comp_op: int
star_expr: int
expr: int
xor_expr: int
and_expr: int
shift_expr: int
arith_expr: int
term: int
factor: int
power: int
atom_expr: int
atom: int
testlist_comp: int
trailer: int
subscriptlist: int
subscript: int
sliceop: int
exprlist: int
testlist: int
dictorsetmaker: int
classdef: int
arglist: int
argument: int
comp_iter: int
comp_for: int
comp_if: int
encoding_decl: int
yield_expr: int
yield_arg: int

sym_name: dict[int, str]



"""Code and data structures for managing source directives."""

import bisect
import collections
import io
import keyword
import logging
import re
import sys
import tokenize
from typing import Collection

from pytype import blocks
from pytype import utils

log = logging.getLogger(__name__)

_DIRECTIVE_RE = re.compile(r"#\s*(pytype|type)\s*:\s?([^#]*)")
_IGNORE_RE = re.compile(r"^ignore(\[.+\])?$")
_CLASS_OR_FUNC_RE = re.compile(r"^(def|class)\s")
_DOCSTRING_RE = re.compile(r"^\s*(\"\"\"|''')")
_DECORATOR_RE = re.compile(r"^\s*@(\w+)([(]|\s*$)")
_ALL_ERRORS = "*"  # Wildcard for disabling all errors.

_FUNCTION_CALL_ERRORS = (
    "duplicate-keyword",
    "missing-parameter",
    "wrong-arg-count",
    "wrong-arg-types",
    "wrong-keyword-args",
)


class _DirectiveError(Exception):
  pass


class SkipFileError(Exception):
  """Exception thrown if we encounter "pytype: skip-file" in the source code."""


class _LineSet:
  """A set of line numbers.

  The data structure is optimized to represent the union of a sparse set
  of integers and ranges of non-negative integers.  This supports the two styles
  of directives: those after a statement apply only to that line and those on
  their own line apply until countered by the opposing directive.
  """

  def __init__(self):
    # Map of line->bool for specific lines, takes precedence over _transitions.
    self._lines = {}
    # A sorted list of the lines at which the range state changes
    # polarity.  It is assumed to initially be false (not in a range).
    # Even positions represent the start of a range, odd positions represent
    # the end of a range.  Thus [2, 5, 10, 12] would include lines 2, 3, 4, 10,
    # and 11.  If the length is odd, then an end of maxint is implied, thus
    # [2, 5, 10] would disable lines 2, 3, 4, 10, 11, 12, ...
    self._transitions = []

  @property
  def lines(self):
    return self._lines

  def set_line(self, line, membership):
    """Set whether a given line is a member of the set."""
    self._lines[line] = membership

  def start_range(self, line, membership):
    """Start a range of lines that are either included/excluded from the set.

    Args:
      line: A line number.
      membership: If True, lines >= line are included in the set (starting
        a range), otherwise they are excluded (ending a range).

    Raises:
      ValueError: if line is less than that of a previous call to start_range().
    """
    last = self._transitions[-1] if self._transitions else -1
    # Assert that lines are monotonically increasing.  This simplifies the
    # logic of adding new lines and ensures that _ranges is sorted.
    if line < last:
      raise ValueError("Line number less than previous start_range() call.")
    # Determine previous membership state (True if the last range has an
    # indefinite end).
    previous = (len(self._transitions) % 2) == 1
    if membership == previous:
      return  # Redundant with previous state, do nothing.
    elif line == last:
      # We have either enable/disable or disable/enable on the same line,
      # cancel them out by popping the previous transition.
      self._transitions.pop()
    else:
      # Normal case - add a transition at this line.
      self._transitions.append(line)

  def __contains__(self, line):
    """Return if a line is a member of the set."""
    # First check for an entry in _lines.
    specific = self._lines.get(line)
    if specific is not None:
      return specific
    # Find the position in _ranges for line.  The polarity of this position
    # determines whether we are inside a range (odd) or outside (even).
    pos = bisect.bisect(self._transitions, line)
    return (pos % 2) == 1

  def get_disable_after(self, lineno):
    """Get an unclosed disable, if any, that starts after lineno."""
    if len(self._transitions) % 2 == 1 and self._transitions[-1] >= lineno:
      return self._transitions[-1]
    return None


class _TypeCommentSet:
  """A set of type comments in a single logical line."""

  @classmethod
  def start(cls, lineno):
    return cls(lineno)

  def __init__(self, start_lineno):
    self.start_line = start_lineno
    self.end_line = None
    self.type_comments = {}


class _FunctionDefinition:
  """Tracks the line numbers of function definitions."""

  @classmethod
  def start(cls, lineno):
    return cls(lineno)

  def __init__(self, start_lineno):
    self._paren_count = 0
    self._start_line = start_lineno
    self._end_line = None

  def add_lpar(self, lineno):
    assert lineno >= self._start_line
    self._paren_count += 1

  def add_rpar(self, lineno):
    if self._end_line is not None:
      return
    self._paren_count -= 1
    if self._paren_count == 0:
      self._end_line = lineno

  def contains(self, lineno):
    if lineno < self._start_line:
      return False
    return self._end_line is None or lineno <= self._end_line


class _VariableAnnotation:
  """Processes a single logical line, looking for a variable annotation."""

  @classmethod
  def start(cls, lineno, token):
    self = cls()
    self.add_token(lineno, token)
    return self

  def __init__(self):
    self._tokens = []
    self.annotation = ""
    # Set to True when the full annotation has been found, or if we determine
    # that the line does not contain an annotation.
    self.closed = False
    # Set to the line on which the colon is found. We do not use the line at
    # which start() is called because the latter may be a blank line.
    self.start_lineno = None
    # Used to consume a 'self.' or 'cls.' prefix so we can detect variable
    # annotations on attributes.
    self._attr_prefix = []

  def _add_to_attr_prefix(self, token):
    if self._tokens:
      return False
    if (not self._attr_prefix and token.string in ("self", "cls") or
        len(self._attr_prefix) == 1 and token.exact_type == tokenize.DOT):
      self._attr_prefix.append(token)
      return True
    return False

  def _accept(self, token):
    if self.closed:
      return False
    # Allow comments and whitespace before the NAME token signifying the start
    # of the annotation.
    return token.exact_type != tokenize.COMMENT and token.string.strip()

  def add_token(self, lineno, token):
    """Process a token."""
    if not self._accept(token):
      return
    if self._add_to_attr_prefix(token):
      return
    # Match NAME COLON [annotation] EQUAL. We assume the annotation starts at
    # the beginning of the line, which greatly simplifies matching at the cost
    # of failing to find annotations in lines like `if __random__: v: int = 0`.
    if not self._tokens:
      # Filter out false positives like `else: x = 0`.
      if token.exact_type != tokenize.NAME or keyword.iskeyword(token.string):
        self.closed = True
    elif len(self._tokens) == 1:
      if token.exact_type == tokenize.COLON:
        self.start_lineno = lineno
      else:
        self.closed = True
    elif token.exact_type == tokenize.EQUAL:
      self.closed = True
    else:
      if self.annotation and self._tokens[-1].end[0] == token.start[0]:
        # Preserve whitespace.
        self.annotation += token.line[self._tokens[-1].end[1]:token.start[1]]
      self.annotation += token.string
    self._tokens.append(token)


def _collect_bytecode(ordered_code):
  bytecode_blocks = []
  stack = [ordered_code]
  while stack:
    code = stack.pop()
    bytecode_blocks.append(code.original_co_code)
    for const in code.co_consts:
      if isinstance(const, blocks.OrderedCode):
        stack.append(const)
  return bytecode_blocks


def _adjust_line_number(line, allowed_lines, min_line=1):
  adjusted_line = line
  while adjusted_line not in allowed_lines and adjusted_line >= min_line:
    adjusted_line -= 1
  return adjusted_line if adjusted_line >= min_line else None


def _is_function_call(opcode_name):
  return opcode_name.startswith("CALL_") or opcode_name in {
      "BINARY_SUBSCR",
      "COMPARE_OP",
      "FOR_ITER",
  }


def _is_funcdef_op(opcode_name):
  """Checks whether the opcode may appear in a function signature."""
  return (opcode_name.startswith("BUILD_") or
          opcode_name.startswith("LOAD_") or
          opcode_name == "BINARY_SUBSCR")


def _is_load_attribute_op(opcode_name):
  """Checks whether the opcode loads an attribute."""
  return (opcode_name.startswith("GET_") or
          opcode_name.startswith("UNPACK_") or
          opcode_name in {
              "LOAD_ATTR",
              "LOAD_METHOD",
              "SETUP_WITH",
          })


def _is_return_op(opcode_name):
  return opcode_name.startswith("YIELD_") or opcode_name == "RETURN_VALUE"


class _OpcodeLines:
  """Stores opcode line numbers for Director.adjust_line_numbers()."""

  class Call:

    def __init__(self, line):
      self.line = line
      self.children = set()

    def __repr__(self):
      return f"call @ line {self.line} ({len(self.children)} nested calls)"

  def __init__(
      self,
      store_lines: Collection[int],
      make_function_lines: Collection[int],
      non_funcdef_lines: Collection[int],
      load_attr_lines: Collection[int],
      return_lines: Collection[int],
      call_lines: Collection[Call]):
    self.store_lines = store_lines
    self.make_function_lines = make_function_lines
    self.non_funcdef_lines = non_funcdef_lines
    self.load_attr_lines = load_attr_lines
    self.return_lines = return_lines
    # We transform call_lines into a line->Call mapping so that
    # _adjust_line_number can treat it as a Collection[int] like the other
    # *_lines attributes.
    self.call_lines = {call.line: call for call in call_lines}

  @classmethod
  def from_code(cls, code):
    """Builds an _OpcodeLines from a code object."""
    store_lines = set()
    make_function_lines = set()
    non_funcdef_lines = set()
    load_attr_lines = set()
    return_lines = set()
    all_call_lines = []
    for block in _collect_bytecode(code):
      call_lines = []
      for opcode in block:
        if opcode.name.startswith("STORE_"):
          store_lines.add(opcode.line)
        elif opcode.name == "MAKE_FUNCTION":
          make_function_lines.add(opcode.line)
        elif _is_load_attribute_op(opcode.name):
          load_attr_lines.add(opcode.line)
        elif _is_return_op(opcode.name):
          return_lines.add(opcode.line)
        elif _is_function_call(opcode.name):
          # Function calls can be nested, so we represent them as a sequence of
          # call trees. As opcodes are typically ordered by increasing line
          # number, we detect nested calls via decreasing line numbers. For
          # example, for:
          #   top_level1()  # line 1
          #   top_level2(  # line 2
          #     nested1(  # line 3
          #       more_nested()),  # line 4
          #     nested2())  # line 5
          # the sequence of line numbers is [1, 4, 3, 5, 2].
          call = _OpcodeLines.Call(opcode.line)
          while call_lines and opcode.line < call_lines[-1].line:
            call.children.add(call_lines.pop())
          call_lines.append(call)
        if not _is_funcdef_op(opcode.name):
          non_funcdef_lines.add(opcode.line)
      all_call_lines.extend(call_lines)
    return cls(store_lines, make_function_lines, non_funcdef_lines,
               load_attr_lines, return_lines, all_call_lines)


class Director:
  """Holds all of the directive information for a source file."""

  def __init__(self, src, errorlog, filename, disable, python_version):
    """Create a Director for a source file.

    Args:
      src:  The source text as a string.
      errorlog: An ErrorLog object.  Directive errors will be logged to the
          errorlog.
      filename: The name of the source file.
      disable: List of error messages to always ignore.
      python_version: The target python version.
    """
    self._filename = filename
    self._errorlog = errorlog
    self._type_comments = []  # _TypeCommentSet objects.
    self._variable_annotations = {}  # Map from line number to annotation.
    self._docstrings = set()  # Start lines of docstrings.
    # Lines that have "type: ignore".  These will disable all errors, and in
    # the future may have other impact (such as not attempting an import).
    self._ignore = _LineSet()
    # Map from error name to lines for which that error is disabled.  Note
    # that _ALL_ERRORS is essentially a wildcard name (it matches all names).
    self._disables = collections.defaultdict(_LineSet)
    # Line numbers of decorators. Since this is used to mark a class or function
    # as decorated, stacked decorators will record the one closest to the
    # definition (i.e. the last one). The python bytecode uses this line number
    # for all the stacked decorator invocations.
    self._decorators = set()
    # Apply global disable, from the command line arguments:
    for error_name in disable:
      self._disables[error_name].start_range(0, True)
    # Parse the source code for directives.
    self._parse_source(src, python_version)

  @property
  def type_comments(self):
    return collections.ChainMap(*(s.type_comments for s in self._type_comments))

  @property
  def annotations(self):
    # It's okay to overwrite type comments with variable annotations here
    # because _FindIgnoredTypeComments in vm.py will flag ignored comments.
    return {**self.type_comments, **self._variable_annotations}

  @property
  def docstrings(self):
    return sorted(self._docstrings)

  @property
  def ignore(self):
    return self._ignore

  @property
  def decorators(self):
    return self._decorators

  def _parse_source(self, src, python_version):
    """Parse a source file, extracting directives from comments."""
    f = io.StringIO(src)
    defs_start = None
    open_type_comment_set = _TypeCommentSet.start(1)
    open_decorator = False
    last_function_definition = None
    open_variable_annotation = None
    for token in tokenize.generate_tokens(f.readline):
      tok = token.exact_type
      line = token.line
      lineno, col = token.start

      # Check for the first line with a top-level class or function definition.
      if defs_start is None and _CLASS_OR_FUNC_RE.match(line):
        defs_start = lineno

      # Process the token for decorators, function definitions, and comments.
      if tok == tokenize.AT:
        if _DECORATOR_RE.match(line):
          open_decorator = True
      elif tok == tokenize.NAME:
        if open_decorator and token.string in ("class", "def"):
          self.decorators.add(lineno)
          open_decorator = False
        if token.string == "def":
          last_function_definition = _FunctionDefinition.start(lineno)
      elif tok == tokenize.COMMENT:
        self._process_comment(line, lineno, col, open_type_comment_set)
      elif tok == tokenize.LPAR:
        if last_function_definition:
          last_function_definition.add_lpar(lineno)
      elif tok == tokenize.RPAR:
        if last_function_definition:
          last_function_definition.add_rpar(lineno)
      elif tok in (tokenize.NEWLINE, tokenize.ENDMARKER):
        if open_type_comment_set.type_comments:
          open_type_comment_set.end_line = lineno
          self._type_comments.append(open_type_comment_set)
        open_type_comment_set = _TypeCommentSet.start(lineno + 1)

      # Process the token for variable annotations.
      if last_function_definition and last_function_definition.contains(lineno):
        pass  # ignore function annotations
      elif not open_variable_annotation:
        open_variable_annotation = _VariableAnnotation.start(lineno, token)
      elif tok in (tokenize.NEWLINE, tokenize.SEMI):
        # NEWLINE indicates the end of a *logical* line of Python code, allowing
        # us to handle annotations split over multiple lines.
        annotation = open_variable_annotation.annotation
        if annotation and open_variable_annotation.closed:
          # In 3.8+, the STORE_* opcode for a multiline variable assignment is
          # at the first line in the assignment; before that, it is at the last.
          if python_version >= (3, 8):
            assert open_variable_annotation.start_lineno
            annotation_lineno = open_variable_annotation.start_lineno
          else:
            annotation_lineno = lineno
          self._variable_annotations[annotation_lineno] = annotation
        open_variable_annotation = None
      else:
        open_variable_annotation.add_token(lineno, token)

      # Record docstrings.
      if _DOCSTRING_RE.match(line):
        self._docstrings.add(lineno)

    if defs_start is not None:
      disables = list(self._disables.items())
      # Add "# type: ignore" to the list of disables that we check.
      disables.append(("Type checking", self._ignore))
      for name, lineset in disables:
        lineno = lineset.get_disable_after(defs_start)
        if lineno is not None:
          self._errorlog.late_directive(self._filename, lineno, name)

  def _process_comment(self, line, lineno, col, type_comment_set):
    """Process a single comment."""
    matches = list(_DIRECTIVE_RE.finditer(line[col:]))
    is_nested = bool(matches) and matches[0].start(0) > 0
    for m in matches:
      code = line[:col].strip()
      open_ended = not code
      tool, data = m.groups()
      assert data is not None
      data = data.strip()
      if tool == "type":
        self._process_type(lineno, code, data, is_nested, type_comment_set)
      elif tool == "pytype":
        try:
          self._process_pytype(lineno, data, open_ended)
        except _DirectiveError as e:
          self._errorlog.invalid_directive(
              self._filename, lineno, utils.message(e))
      else:
        pass  # ignore comments for other tools

  def _process_type(self, lineno, code, data, is_nested, type_comment_set):
    """Process a type: comment."""
    # Discard type comments embedded in larger whole-line comments.
    if not code and is_nested:
      return
    if lineno in type_comment_set.type_comments:
      # If we have multiple type comments on the same line, take the last one,
      # but add an error to the log.
      self._errorlog.invalid_directive(
          self._filename, lineno,
          "Multiple type comments on the same line.")
    # Also supports mypy-style ignore[code, ...] syntax, treated as regular
    # ignores.
    if _IGNORE_RE.match(data):
      if not code:
        self._ignore.start_range(lineno, True)
      else:
        self._ignore.set_line(lineno, True)
    else:
      type_comment_set.type_comments[lineno] = data

  def _process_pytype(self, lineno, data, open_ended):
    """Process a pytype: comment."""
    if not data:
      raise _DirectiveError("Invalid directive syntax.")
    for option in data.split():
      # Parse the command.
      if option == "skip-file":
        raise SkipFileError()
      try:
        command, values = option.split("=", 1)
        values = values.split(",")
      except ValueError as e:
        raise _DirectiveError("Invalid directive syntax.") from e
      # Additional commands may be added in the future.  For now, only
      # "disable" and "enable" are supported.
      if command == "disable":
        disable = True
      elif command == "enable":
        disable = False
      else:
        raise _DirectiveError("Unknown pytype directive: '%s'" % command)
      if not values:
        raise _DirectiveError(
            "Disable/enable must specify one or more error names.")
      for error_name in values:
        if (error_name == _ALL_ERRORS or
            self._errorlog.is_valid_error_name(error_name)):
          lines = self._disables[error_name]
          if open_ended:
            lines.start_range(lineno, disable)
          else:
            lines.set_line(lineno, disable)
        else:
          self._errorlog.invalid_directive(
              self._filename, lineno, "Invalid error name: '%s'" % error_name)

  def should_report_error(self, error):
    """Return whether the error should be logged.

    This method is suitable for use as an error filter.

    Args:
      error: An error._Error object.

    Returns:
      True iff the error should be included in the log.
    """
    # Always report errors that aren't for this file or do not have a line
    # number.
    if error.filename != self._filename or error.lineno is None:
      return True
    # Treat lineno=0 as below the file, so we can filter it.
    lineno = error.lineno or sys.maxsize
    # Report the error if it isn't subject to any ignore or disable.
    return (lineno not in self._ignore and
            lineno not in self._disables[_ALL_ERRORS] and
            lineno not in self._disables[error.name])

  def _adjust_line_numbers_for_type_comments(self, opcode_lines):
    """Adjusts line numbers for `# type:` comments."""
    for type_comment_set in self._type_comments:
      for line, comment in sorted(type_comment_set.type_comments.items()):
        adjusted_line = _adjust_line_number(
            line, opcode_lines.store_lines, type_comment_set.start_line)
        if not adjusted_line:
          # vm._FindIgnoredTypeComments will take care of error reporting.
          continue
        if line != type_comment_set.end_line:
          self._errorlog.ignored_type_comment(self._filename, line, comment)
          del type_comment_set.type_comments[line]
        elif adjusted_line != line:
          type_comment_set.type_comments[adjusted_line] = comment
          del type_comment_set.type_comments[line]

  def _adjust_line_numbers_for_decorators(self, opcode_lines):
    for line in sorted(self._decorators):
      adjusted_line = _adjust_line_number(
          line, opcode_lines.make_function_lines)
      if not adjusted_line:
        log.warning(
            "No MAKE_FUNCTION opcode found for decorator on line %d", line)
      elif adjusted_line != line:
        self._decorators.add(adjusted_line)
        self._decorators.remove(line)

  def _adjust_line_numbers_for_variable_annotations(self, opcode_lines):
    for line, annot in sorted(self._variable_annotations.items()):
      adjusted_line = _adjust_line_number(line, opcode_lines.store_lines)
      if not adjusted_line:
        log.error(
            "No STORE_* opcode found for annotation %r on line %d", annot, line)
        del self._variable_annotations[line]
      elif adjusted_line != line:
        self._variable_annotations[adjusted_line] = annot
        del self._variable_annotations[line]

  def _adjust_line_numbers_for_error_directives(self, opcode_lines):
    """Adjusts line numbers for error directives."""
    for error_class in _FUNCTION_CALL_ERRORS + (
        "annotation-type-mismatch", "attribute-error", "bad-return-type"):
      if error_class not in self._disables:
        continue
      lines = self._disables[error_class].lines
      for line, membership in sorted(lines.items()):
        if error_class == "annotation-type-mismatch":
          min_line = line
          # In Python 3.8+, the MAKE_FUNCTION opcode's line number is the first
          # line of the function signature, so we need to skip any opcodes
          # associated with the signature in between.
          while min_line not in opcode_lines.non_funcdef_lines and min_line > 1:
            min_line -= 1
          allowed_lines = (
              opcode_lines.store_lines | opcode_lines.make_function_lines)
        elif error_class == "attribute-error":
          min_line = 1
          allowed_lines = opcode_lines.load_attr_lines
        elif error_class == "bad-return-type":
          min_line = 1
          allowed_lines = opcode_lines.return_lines
        else:
          min_line = 1
          allowed_lines = opcode_lines.call_lines
        adjusted_line = _adjust_line_number(line, allowed_lines, min_line)
        if adjusted_line and adjusted_line != line:
          if error_class in (
              "annotation-type-mismatch", "attribute-error", "bad-return-type"):
            lines[adjusted_line] = membership
            del lines[line]
          else:
            # 'adjusted_line' points to the top-level function call that
            # contains 'line'. We apply the directive on 'line' to the top-level
            # call as well as all nested calls between 'adjusted_line' and
            # 'line'. For example, in:
            # top_level_call(
            #     nested1(),
            #     nested2(),  # directive
            #     nested3())
            # the disable directive applies to top_level_call, nested1, and
            # nested 2, but not nested3.
            stack = [opcode_lines.call_lines[adjusted_line]]
            found = False
            while stack:
              call = stack.pop()
              if call.line > line:
                continue
              if call.line == line:
                found = True
              else:
                lines[call.line] = membership
              stack.extend(call.children)
            # If 'line' does not itself contain a function call, then we delete
            # the directive on 'line' after applying it to all the relevant
            # lines that do contain function calls.
            if not found:
              del lines[line]

  def adjust_line_numbers(self, code):
    """Uses the bytecode to adjust line numbers."""
    opcode_lines = _OpcodeLines.from_code(code)
    self._adjust_line_numbers_for_type_comments(opcode_lines)
    self._adjust_line_numbers_for_decorators(opcode_lines)
    self._adjust_line_numbers_for_variable_annotations(opcode_lines)
    self._adjust_line_numbers_for_error_directives(opcode_lines)



from abc import abstractmethod
from collections.abc import Iterable, Mapping
from distutils.core import Command as _Command
from typing import Any, Type

from setuptools._deprecation_warning import SetuptoolsDeprecationWarning as SetuptoolsDeprecationWarning
from setuptools.depends import Require as Require
from setuptools.dist import Distribution as Distribution
from setuptools.extension import Extension as Extension

__version__: str

class PackageFinder:
    @classmethod
    def find(cls, where: str = ..., exclude: Iterable[str] = ..., include: Iterable[str] = ...) -> list[str]: ...

class PEP420PackageFinder(PackageFinder): ...

find_packages = PackageFinder.find
find_namespace_packages = PEP420PackageFinder.find

def setup(
    *,
    name: str = ...,
    version: str = ...,
    description: str = ...,
    long_description: str = ...,
    author: str = ...,
    author_email: str = ...,
    maintainer: str = ...,
    maintainer_email: str = ...,
    url: str = ...,
    download_url: str = ...,
    packages: list[str] = ...,
    py_modules: list[str] = ...,
    scripts: list[str] = ...,
    ext_modules: list[Extension] = ...,
    classifiers: list[str] = ...,
    distclass: Type[Distribution] = ...,
    script_name: str = ...,
    script_args: list[str] = ...,
    options: Mapping[str, Any] = ...,
    license: str = ...,
    keywords: list[str] | str = ...,
    platforms: list[str] | str = ...,
    cmdclass: Mapping[str, Type[Command]] = ...,
    data_files: list[tuple[str, list[str]]] = ...,
    package_dir: Mapping[str, str] = ...,
    obsoletes: list[str] = ...,
    provides: list[str] = ...,
    requires: list[str] = ...,
    command_packages: list[str] = ...,
    command_options: Mapping[str, Mapping[str, tuple[Any, Any]]] = ...,
    package_data: Mapping[str, list[str]] = ...,
    include_package_data: bool = ...,
    libraries: list[str] = ...,
    headers: list[str] = ...,
    ext_package: str = ...,
    include_dirs: list[str] = ...,
    password: str = ...,
    fullname: str = ...,
    **attrs: Any,
) -> None: ...

class Command(_Command):
    command_consumes_arguments: bool
    def __init__(self, dist: Distribution, **kw: Any) -> None: ...
    def ensure_string_list(self, option: str | list[str]) -> None: ...
    def reinitialize_command(self, command: _Command | str, reinit_subcommands: int = ..., **kw: Any) -> _Command: ...
    @abstractmethod
    def initialize_options(self) -> None: ...
    @abstractmethod
    def finalize_options(self) -> None: ...
    @abstractmethod
    def run(self) -> None: ...

class sic(str): ...



"""Base value for abstract classes.

Unless forced to by a circular dependency, don't import BaseValue directly from
this module; use the alias in abstract.py instead.
"""

import hashlib
from typing import Any

from pytype import utils
from pytype.abstract import abstract_utils
from pytype.pytd import mro

_isinstance = abstract_utils._isinstance  # pylint: disable=protected-access
_make = abstract_utils._make  # pylint: disable=protected-access


class BaseValue(utils.ContextWeakrefMixin):
  """A single abstract value such as a type or function signature.

  This is the base class of the things that appear in Variables. It represents
  an atomic object that the abstract interpreter works over just as variables
  represent sets of parallel options.

  Conceptually abstract values represent sets of possible concrete values in
  compact form. For instance, an abstract value with .__class__ = int represents
  all ints.
  """

  formal = False  # is this type non-instantiable?

  def __init__(self, name, ctx):
    """Basic initializer for all BaseValues."""
    super().__init__(ctx)
    # This default cls value is used by things like Unsolvable that inherit
    # directly from BaseValue. Classes and instances overwrite the default with
    # a more sensible value.
    self.cls = self
    self.name = name
    self.mro = self.compute_mro()
    self.module = None
    self.official_name = None
    self.slots = None  # writable attributes (or None if everything is writable)
    # true for functions and classes that have decorators applied to them.
    self.is_decorated = False
    # The template for the current class. It is usually a constant, lazily
    # loaded to accommodate recursive types, but in the case of typing.Generic
    # (only), it'll change every time when a new generic class is instantiated.
    self._template = None
    # names in the templates of the current class and its base classes
    self._all_template_names = None
    self._instance = None

    # The variable or function arg name with the type annotation that this
    # instance was created from. For example,
    #   x: str = "hello"
    # would create an instance of str with from_annotation = 'x'
    self.from_annotation = None

  @property
  def all_template_names(self):
    if self._all_template_names is None:
      self._all_template_names = _get_template(self)
    return self._all_template_names

  @property
  def template(self):
    if self._template is None:
      # Won't recompute if `compute_template` throws exception
      self._template = ()
      self._template = _compute_template(self)
    return self._template

  @property
  def full_name(self):
    return (self.module + "." if self.module else "") + self.name

  def __repr__(self):
    return self.name

  def compute_mro(self):
    # default for objects with no MRO
    return []

  def default_mro(self):
    # default for objects with unknown MRO
    return [self, self.ctx.convert.object_type]

  def get_fullhash(self):
    """Hash this value and all of its children."""
    m = hashlib.md5()
    seen_data = set()
    stack = [self]
    while stack:
      data = stack.pop()
      data_hash = hash(data)
      if data_hash in seen_data:
        continue
      seen_data.add(data_hash)
      m.update(str(data_hash).encode("utf-8"))
      for mapping in data.get_children_maps():
        m.update(str(mapping.changestamp).encode("utf-8"))
        stack.extend(mapping.data)
    return m.digest()

  def get_children_maps(self):
    """Get this value's dictionaries of children.

    Returns:
      A sequence of dictionaries from names to child values.
    """
    return ()

  def get_instance_type_parameter(self, name, node=None):
    """Get a cfg.Variable of the instance's values for the type parameter.

    Treating self as an abstract.Instance, gets the variable of its values for
    the given type parameter. For the real implementation, see
    SimpleValue.get_instance_type_parameter.

    Args:
      name: The name of the type parameter.
      node: Optionally, the current CFG node.
    Returns:
      A Variable which may be empty.
    """
    del name
    if node is None:
      node = self.ctx.root_node
    return self.ctx.new_unsolvable(node)

  def get_formal_type_parameter(self, t):
    """Get the class's type for the type parameter.

    Treating self as a class_mixin.Class, gets its formal type for the given
    type parameter. For the real implementation, see
    ParameterizedClass.get_formal_type_parameter.

    Args:
      t: The name of the type parameter.
    Returns:
      A formal type.
    """
    del t
    return self.ctx.convert.unsolvable

  def property_get(self, callself, is_class=False):
    """Bind this value to the given self or cls.

    This function is similar to __get__ except at the abstract level. This does
    not trigger any code execution inside the VM. See __get__ for more details.

    Args:
      callself: The Variable that should be passed as self or cls when the call
        is made. We only need one of self or cls, so having them share a
        parameter prevents accidentally passing in both.
      is_class: Whether callself is self or cls. Should be cls only when we
        want to directly pass in a class to bind a class method to, rather than
        passing in an instance and calling get_class().

    Returns:
      Another abstract value that should be returned in place of this one. The
      default implementation returns self, so this can always be called safely.
    """
    del callself, is_class
    return self

  def get_special_attribute(self, unused_node, name, unused_valself):
    """Fetch a special attribute (e.g., __get__, __iter__)."""
    if name == "__class__":
      if self.full_name == "typing.Protocol":
        # Protocol.__class__ is a _ProtocolMeta class that inherits from
        # abc.ABCMeta. Changing the definition of Protocol in typing.pytd to
        # include this metaclass causes a bunch of weird breakages, so we
        # instead return the metaclass when type() or __class__ is accessed on
        # Protocol. For simplicity, we pretend the metaclass is ABCMeta rather
        # than a subclass.
        abc = self.ctx.vm.import_module("abc", "abc", 0).get_module("ABCMeta")
        abc.load_lazy_attribute("ABCMeta")
        return abc.members["ABCMeta"]
      else:
        return self.cls.to_variable(self.ctx.root_node)
    return None

  def get_own_new(self, node, value):
    """Get this value's __new__ method, if it isn't object.__new__."""
    del value  # Unused, only classes have methods.
    return node, None

  def call(self, node, func, args, alias_map=None):
    """Call this abstract value with the given arguments.

    The posargs and namedargs arguments may be modified by this function.

    Args:
      node: The CFGNode calling this function
      func: The cfg.Binding containing this function.
      args: Arguments for the call.
      alias_map: A datatypes.UnionFind, which stores all the type renaming
        information, mapping of type parameter name to its representative.
    Returns:
      A tuple (cfg.Node, cfg.Variable). The CFGNode corresponds
      to the function's "return" statement(s).
    Raises:
      function.FailedFunctionCall

    Make the call as required by this specific kind of atomic value, and make
    sure to annotate the results correctly with the origins (val and also other
    values appearing in the arguments).
    """
    raise NotImplementedError(self.__class__.__name__)

  def argcount(self, node):
    """Returns the minimum number of arguments needed for a call."""
    raise NotImplementedError(self.__class__.__name__)

  def register_instance(self, instance):  # pylint: disable=unused-arg
    """Treating self as a class definition, register an instance of it.

    This is used for keeping merging call records on instances when generating
    the formal definition of a class. See InterpreterClass and TupleClass.

    Args:
      instance: An instance of this class (as a BaseValue)
    """

  def get_instance_type(self, node=None, instance=None, seen=None, view=None):
    """Get the type an instance of us would have."""
    return self.ctx.pytd_convert.value_instance_to_pytd_type(
        node, self, instance, seen, view)

  def to_type(self, node=None, seen=None, view=None):
    """Get a PyTD type representing this object, as seen at a node."""
    return self.ctx.pytd_convert.value_to_pytd_type(node, self, seen, view)

  def to_pytd_def(self, node, name):
    """Get a PyTD definition for this object."""
    return self.ctx.pytd_convert.value_to_pytd_def(node, self, name)

  def get_default_type_key(self):
    """Gets a default type key. See get_type_key."""
    return type(self)

  def get_type_key(self, seen=None):  # pylint: disable=unused-argument
    """Build a key from the information used to perform type matching.

    Get a hashable object containing this value's type information. Type keys
    are only compared amongst themselves, so we don't care what the internals
    look like, only that values with different types *always* have different
    type keys and values with the same type preferably have the same type key.

    Args:
      seen: The set of values seen before while computing the type key.

    Returns:
      A hashable object built from this value's type information.
    """
    return self.get_default_type_key()

  def instantiate(self, node, container=None):
    """Create an instance of self.

    Note that this method does not call __init__, so the instance may be
    incomplete. If you need a complete instance, use self.ctx.vm.init_class
    instead.

    Args:
      node: The current node.
      container: Optionally, the value that contains self. (See TypeParameter.)

    Returns:
      The instance.
    """
    raise NotImplementedError(self.__class__.__name__)

  def to_annotation_container(self):
    if _isinstance(self, "PyTDClass") and self.full_name == "builtins.tuple":
      # If we are parameterizing builtins.tuple, replace it with typing.Tuple so
      # that heterogeneous tuple annotations work. We need the isinstance()
      # check to distinguish PyTDClass(tuple) from ParameterizedClass(tuple);
      # the latter appears here when a generic type alias is being substituted.
      typing = self.ctx.vm.import_module("typing", "typing",
                                         0).get_module("Tuple")
      typing.load_lazy_attribute("Tuple")
      return abstract_utils.get_atomic_value(typing.members["Tuple"])
    return _make("AnnotationContainer", self.name, self.ctx, self)

  def to_variable(self, node):
    """Build a variable out of this abstract value.

    Args:
      node: The current CFG node.
    Returns:
      A cfg.Variable.
    """
    return self.ctx.program.NewVariable([self], source_set=[], where=node)

  def to_binding(self, node):
    binding, = self.to_variable(node).bindings
    return binding

  def has_varargs(self):
    """Return True if this is a function and has a *args parameter."""
    return False

  def has_kwargs(self):
    """Return True if this is a function and has a **kwargs parameter."""
    return False

  def _unique_parameters(self):
    """Get unique parameter subtypes as variables.

    This will retrieve 'children' of this value that contribute to the
    type of it. So it will retrieve type parameters, but not attributes. To
    keep the number of possible combinations reasonable, when we encounter
    multiple instances of the same type, we include only one.

    Returns:
      A list of variables.
    """
    return []

  def unique_parameter_values(self):
    """Get unique parameter subtypes as bindings.

    Like _unique_parameters, but returns bindings instead of variables.

    Returns:
      A list of list of bindings.
    """
    def _get_values(parameter):
      return {b.data.get_type_key(): b for b in parameter.bindings}.values()
    return [_get_values(parameter) for parameter in self._unique_parameters()]

  def init_subclass(self, node, cls):
    """Allow metaprogramming via __init_subclass__.

    We do not analyse __init_subclass__ methods in the code, but overlays that
    wish to replicate metaprogramming constructs using __init_subclass__ can
    define a class overriding this method, and ctx.make_class will call
    Class.call_init_subclass(), which will invoke the init_subclass() method for
    all classes in the list of base classes.

    This is here rather than in class_mixin.Class because a class's list of
    bases can include abstract objects that do not derive from Class (e.g.
    Unknown and Unsolvable).

    Args:
      node: cfg node
      cls: the abstract.InterpreterClass that is being constructed with subclass
           as a base
    Returns:
      A possibly new cfg node
    """
    del cls
    return node

  def update_official_name(self, _):
    """Update the official name."""

  def is_late_annotation(self):
    return False


def _get_template(val: Any):
  """Get the value's class template."""
  if _isinstance(val, "Class"):
    res = {t.full_name for t in val.template}
    if _isinstance(val, "ParameterizedClass"):
      res.update(_get_template(val.base_cls))
    elif _isinstance(val, ("PyTDClass", "InterpreterClass")):
      for base in val.bases():
        base = abstract_utils.get_atomic_value(
            base, default=val.ctx.convert.unsolvable)
        res.update(_get_template(base))
    return res
  elif val.cls != val:
    return _get_template(val.cls)
  else:
    return set()


def _compute_template(val: Any):
  """Compute the precedence list of template parameters according to C3.

  1. For the base class list, if it contains `typing.Generic`, then all the
  type parameters should be provided. That means we don't need to parse extra
  base class and then we can get all the type parameters.
  2. If there is no `typing.Generic`, parse the precedence list according to
  C3 based on all the base classes.
  3. If `typing.Generic` exists, it must contain at least one type parameters.
  And there is at most one `typing.Generic` in the base classes. Report error
  if the check fails.

  Args:
    val: The abstract.BaseValue to compute a template for.

  Returns:
    parsed type parameters

  Raises:
    GenericTypeError: if the type annotation for generic type is incorrect
  """
  if _isinstance(val, "PyTDClass"):
    return [
        val.ctx.convert.constant_to_value(itm.type_param)
        for itm in val.pytd_cls.template
    ]
  elif not _isinstance(val, "InterpreterClass"):
    return ()
  bases = [
      abstract_utils.get_atomic_value(base, default=val.ctx.convert.unsolvable)
      for base in val.bases()
  ]
  template = []

  # Compute the number of `typing.Generic` and collect the type parameters
  for base in bases:
    if base.full_name == "typing.Generic":
      if _isinstance(base, "PyTDClass"):
        raise abstract_utils.GenericTypeError(
            val, "Cannot inherit from plain Generic")
      if template:
        raise abstract_utils.GenericTypeError(
            val, "Cannot inherit from Generic[...] multiple times")
      for item in base.template:
        param = base.formal_type_parameters.get(item.name)
        template.append(param.with_module(val.full_name))

  if template:
    # All type parameters in the base classes should appear in
    # `typing.Generic`
    for base in bases:
      if base.full_name != "typing.Generic":
        if _isinstance(base, "ParameterizedClass"):
          for item in base.template:
            param = base.formal_type_parameters.get(item.name)
            if _isinstance(base, "TypeParameter"):
              t = param.with_module(val.full_name)
              if t not in template:
                raise abstract_utils.GenericTypeError(
                    val, "Generic should contain all the type variables")
  else:
    # Compute template parameters according to C3
    seqs = []
    for base in bases:
      if _isinstance(base, "ParameterizedClass"):
        seq = []
        for item in base.template:
          param = base.formal_type_parameters.get(item.name)
          if _isinstance(param, "TypeParameter"):
            seq.append(param.with_module(val.full_name))
        seqs.append(seq)
    try:
      template.extend(mro.MergeSequences(seqs))
    except ValueError as e:
      raise abstract_utils.GenericTypeError(
          val, "Illegal type parameter order in class %s" % val.name) from e

  return template



from typing import Any

from selenium.webdriver.remote.command import Command as Command

from . import interaction as interaction
from .key_actions import KeyActions as KeyActions
from .key_input import KeyInput as KeyInput
from .pointer_actions import PointerActions as PointerActions
from .pointer_input import PointerInput as PointerInput

class ActionBuilder:
    devices: Any
    driver: Any
    def __init__(self, driver, mouse: Any | None = ..., keyboard: Any | None = ...) -> None: ...
    def get_device_with(self, name): ...
    @property
    def pointer_inputs(self): ...
    @property
    def key_inputs(self): ...
    @property
    def key_action(self): ...
    @property
    def pointer_action(self): ...
    def add_key_input(self, name): ...
    def add_pointer_input(self, kind, name): ...
    def perform(self) -> None: ...
    def clear_actions(self) -> None: ...
